# -*- encoding: utf-8 -*-
from functools import wraps

import numpy as np
import pandas as pd


def sample_stats(df, target):
    """
    统计样本中好坏样本数量及坏样本占比，使用示例如下：

    >>> sample_df = pd.DataFrame({
    ...     'id': np.arange(10),
    ...     'prod': ['a', 'a', 'a', 'b', 'b', 'b', 'b', 'c', 'c', 'c'],
    ...     'y': [0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1]})
    >>> sample_df
       id prod  y
    0   0    a  0
    1   1    a  0
    2   2    a  1
    3   3    b  1
    4   4    b  0
    5   5    b  1
    6   6    b  0
    7   7    c  1
    8   8    c  0
    9   9    c  0
    >>> sample_stats(sample_df, target='y')
    TotalCnt     10
    GoodCnt       6
    BadCnt        4
    BadRate     0.4
    dtype: object
    >>> sample_df.groupby('prod').apply(sample_stats, target='y')
          TotalCnt  GoodCnt  BadCnt   BadRate
    prod
    a            3        2       1  0.333333
    b            4        2       2  0.500000
    c            3        2       1  0.333333

    Args:
        df: pd.DataFrame, 样本集
        target: 目标变量名

    Returns:
        pd.Series，各个统计量组成的数列

    """
    y = df[target]
    ttl_cnt = len(y)
    good_cnt = np.sum(y == 0)
    bad_cnt = np.sum(y == 1)
    bad_rate = bad_cnt / (good_cnt + bad_cnt)
    return pd.Series([ttl_cnt, good_cnt, bad_cnt, bad_rate],
                     index=['TotalCnt', 'GoodCnt', 'BadCnt', 'BadRate'],
                     dtype='object')


def monotonic(series: pd.Series) -> str:
    """
    判断一组序列(pd.Series)的单调类型，包括以下五种类型：

    * 'increasing' - 单调递增
    * 'decreasing' - 单调递减
    * 'up_u_shape' - 先单调递减，后单调递增
    * 'down_u_shape' - 先单调递增，后单调递增

    Args:
        series: pd.Series, 数值型数组序列

    Returns:
        str, 单调类型

    """
    if series.is_monotonic_increasing:
        return 'increasing'
    elif series.is_monotonic_decreasing:
        return 'decreasing'
    else:
        for i in range(2, len(series)):
            if series[:i].is_monotonic_increasing and \
                    series[i:].is_monotonic_decreasing:
                return 'down_u_shape'
            elif series[:i].is_monotonic_decreasing and \
                    series[i:].is_monotonic_increasing:
                return 'up_u_shape'
    return 'non_monotonic'


def round_(a, significant_figures=1):
    """
    按照有效数字个数进行四舍五入

    Args:
        a:
        significant_figures:

    Returns:

    """
    e = np.ceil(np.log10(np.where(a == 0, 1, np.abs(a))))
    sig_figs = a / 10**e
    sig_figs = np.round(sig_figs, significant_figures)
    return sig_figs * 10**e


def interactive_mode():
    # noinspection PyPackageRequirements
    import __main__ as main
    if hasattr(main, '__file__'):
        return False
    else:
        return True


def str_to_list(x):
    if x is not None and isinstance(x, str):
        x = [x]
    return x


def exception_trace(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            raise RuntimeError(f'Error in running {func.__name__} with '
                               f'args={args}'
                               f'kwargs={kwargs}')

    return wrapper



# -*- encoding: utf-8 -*-

import numpy as np
import pandas as pd


def make_rule_stat(bad_cnt_fn, good_cnt_fn):

    def rule_stat(df):
        cnt = len(df)
        bad = bad_cnt_fn(df)
        good = good_cnt_fn(df)
        badrate = bad / (bad + good)
        return pd.Series({
            'cnt': cnt,
            'bad': bad,
            'good': good,
            'badrate': badrate
        })

    return rule_stat


def foil_gain(pos0, neg0, pos1, neg1):
    """FOIL增益计算"""
    if pos0 == 0:
        return -np.inf
    if neg0 == 0:
        neg0 += 0.5

    info1 = np.log2(pos0 / (pos0 + neg0))
    info2 = np.log2(pos1 / (pos1 + neg1))

    return pos0 * (info1 - info2)


def eval_rules(df, stat_fn, *rule):
    rules = list(rule)
    metrics = []

    for rule in rules:
        hit_metric = df.groupby(rule(df)).apply(stat_fn)

        if np.any(hit_metric.index):
            hit_rate = hit_metric['cnt'][
                hit_metric.index].item() / hit_metric['cnt'].sum()
            hit_bad = hit_metric['bad'][hit_metric.index].item()
            hit_good = hit_metric['good'][hit_metric.index].item()
            all_bad = hit_metric['bad'].sum()
            all_good = hit_metric['good'].sum()

            if hit_bad + hit_good == 0:
                hit_badrate = 0
                foil = np.inf
            else:
                hit_badrate = hit_bad / (hit_bad + hit_good)
                foil = foil_gain(hit_bad, hit_good, all_bad, all_good)

            metrics.append({
                'rule': rule.__name__,
                'hit': hit_rate,
                'badrate': hit_badrate,
                'foil': foil
            })
        else:
            metrics.append({
                'rule': rule.__name__,
                'hit': 0,
                'badrate': np.nan,
                'foil': 0
            })

    return pd.DataFrame(metrics)


def rule_optimize(df, stat_fn, *rule):
    rules = list(rule)
    named_rules = {rule.__name__: rule for rule in rules}
    opt_result = None
    tmp_df = df.copy()

    while len(tmp_df) > 0 and len(named_rules) > 0:
        rule_result = eval_rules(tmp_df, stat_fn, *named_rules.values())
        best_rule = rule_result.sort_values(
            'foil', ascending=False).iloc[0, ]
        best_rule_name = best_rule['rule']

        if best_rule['foil'] > 0:
            rule = named_rules.pop(best_rule_name)
            tmp_df = tmp_df[~rule(tmp_df)]

            if opt_result is None:
                opt_result = best_rule.to_frame().T

            else:
                opt_result = pd.concat(
                    [opt_result, best_rule.to_frame().T], ignore_index=True)
        else:
            break

    return opt_result



# -*- encoding: utf-8 -*-
import os

import numpy as np
import pandas as pd
import statsmodels.api as sm
from pandas.io.excel import ExcelWriter
from statsmodels.stats.outliers_influence import variance_inflation_factor

from syriskmodels.contrib.var_select import risk_trends_consistency
from syriskmodels.evaluate import (model_eval, gains_table, psi,
                                   swap_analysis_simple)
from syriskmodels.models import group_split_cv, stepwise_lr
from syriskmodels.scorecard import (make_scorecard, sc_bins_to_df, woebin,
                                    woebin_plot, woebin_ply, woebin_psi)
from syriskmodels.utils import str_to_list, sample_stats


def build_scorecard(sample_df,
                    features,
                    target,
                    train_filter,
                    oot_filter,
                    output_excel_file,
                    variable_iv_limit=0.02,
                    sample_stat_group_fields=None,
                    alternative_target=None,
                    regression_p_limit=0.05,
                    random_test_set=0.2,
                    special_values=None,
                    binning_methods=None,
                    binning_kwargs=None,
                    cv=None,
                    base_points=600,
                    base_odds=50,
                    pdo=20,
                    compare_model_fields=None,
                    random_state=0):
    """

    Args:
        sample_df:
        features:
        target:
        train_filter:
        oot_filter:
        output_excel_file:
        variable_iv_limit:
        sample_stat_group_fields:
        alternative_target:
        regression_p_limit:
        random_test_set:
        special_values:
        binning_methods:
        binning_kwargs:
        cv:
        base_points:
        base_odds:
        pdo:
        compare_model_fields:
        random_state:

    Returns:

    """
    if binning_methods is None:
        binning_methods = ['quantile', 'tree']

    if cv is None:
        cv = 3

    # check features
    features = list(set(features).intersection(sample_df.columns.tolist()))

    # check target
    assert target in sample_df.columns.tolist(), f'数据集中不存在目标字段{target}'

    # open excel file
    excel_file = ExcelWriter(output_excel_file)

    # 1. 样本统计
    if sample_stat_group_fields is not None:
        sample_stat_group_fields = str_to_list(sample_stat_group_fields)
        sample_stat_group_fields = list(
            set(sample_stat_group_fields).intersection(
                sample_df.columns.tolist()))

    row_cnt = 0
    if sample_stat_group_fields:
        for field in sample_stat_group_fields:
            tmp_df = sample_df.groupby(field).apply(sample_stats, target=target)
            tmp_df.to_excel(excel_file, sheet_name=f'样本统计', startrow=row_cnt)
            row_cnt += len(tmp_df) + 2

    # 2. 样本拆分
    sample_df['_train_test_flag_'] = np.where(
        train_filter(sample_df), '01_train',
        np.where(oot_filter(sample_df), '03_oot', ''))

    rng = np.random.default_rng(random_state)

    if random_test_set:

        sample_df['_rand_'] = rng.random(len(sample_df))
        sample_df['_train_test_flag_'] = np.where(
            (sample_df['_train_test_flag_'] == '01_train') &
            (sample_df['_rand_'] < random_test_set), '02_test',
            sample_df['_train_test_flag_'])

    tmp_df = sample_df.groupby('_train_test_flag_').apply(sample_stats,
                                                          target=target)
    tmp_df.to_excel(excel_file, sheet_name='样本统计', startrow=row_cnt)

    train_df = sample_df[(sample_df['_train_test_flag_'] == '01_train') &
                         (sample_df[target].isin([0, 1]))].copy().reset_index(
                             drop=True)
    oot_df = sample_df[(sample_df['_train_test_flag_'] == '03_oot') &
                       (sample_df[target].isin([0, 1]))].copy().reset_index(
                           drop=True)

    # 3. 变量分箱
    if binning_kwargs is None:
        binning_kwargs = {}
    bins = woebin(train_df,
                  x=features,
                  y=target,
                  special_values=special_values,
                  methods=binning_methods,
                  **binning_kwargs)
    woe_df, iv_df = sc_bins_to_df(bins)
    woe_df.to_excel(excel_file, sheet_name='WOE分析', index_label='index')
    iv_df.to_excel(excel_file, sheet_name='IV分析', index_label='变量')

    # 变量筛选
    selected_variables = iv_df[
        (iv_df['IV'] > variable_iv_limit) &
        iv_df['单调性'].isin(['increasing', 'decreasing'])].index.tolist()

    var_risk_consist = risk_trends_consistency(
        oot_df, sc_bins={v: bins[v] for v in selected_variables}, target=target)
    selected_variables = [k for k, v in var_risk_consist.items() if v == 1.0]

    # 4.逐步回归
    train_X = woebin_ply(train_df[selected_variables], bins, value='woe')
    train_y = train_df[target]

    _, selected_variables = stepwise_lr(
        train_X,
        train_y.values,
        cv=(lambda: group_split_cv(train_df[cv]))
        if isinstance(cv, str) else cv,
        x=[f + '_woe' for f in selected_variables],
        max_num_features=30)

    print(f'selected variables: {",".join(selected_variables)}')

    # 5. fine tuning
    while True:
        X = train_X[selected_variables].copy()
        X = sm.add_constant(X)
        lr_model = sm.GLM(endog=train_y, exog=X, family=sm.families.Binomial())
        lr_model_result = lr_model.fit()

        if np.any(lr_model_result.params > 0) or np.any(
                lr_model_result.pvalues > regression_p_limit):
            t_values = lr_model_result.tvalues
            rm_var = t_values.index[t_values == t_values.max()].item()
            selected_variables.remove(rm_var)
        else:
            break

    print(selected_variables)
    print(lr_model_result.summary())
    model_summary = pd.DataFrame(
        {'': lr_model_result.summary().as_text().split('\n')})
    model_summary.to_excel(excel_file,
                           sheet_name='模型表达',
                           index=False,
                           header=False)

    # 计算VIF
    vif = {}
    for idx, feature in enumerate(X.columns[1:].tolist()):
        vif[feature] = variance_inflation_factor(train_X, idx)
    vif = pd.Series(vif).rename('VIF').to_frame()
    vif = vif.loc[selected_variables,]
    print(vif)
    vif.to_excel(excel_file,
                 sheet_name='模型表达',
                 index=False,
                 startcol=11,
                 startrow=14)

    scorecard = make_scorecard(bins,
                               lr_model_result.params.to_dict(),
                               base_points=base_points,
                               base_odds=base_odds,
                               pdo=pdo)
    print(scorecard)
    scorecard.to_excel(excel_file,
                       sheet_name='模型表达',
                       index=False,
                       startrow=len(model_summary) + 3)

    # 模型评估
    all_sample_X = woebin_ply(
        sample_df[[var[:-4] for var in selected_variables]], bins)
    # 这一步非常重要！！
    all_sample_X = all_sample_X[selected_variables]
    all_sample_X = sm.add_constant(all_sample_X)

    sample_df['prob'] = lr_model_result.predict(all_sample_X)
    perf = sample_df.groupby('_train_test_flag_').apply(model_eval,
                                                        target=target,
                                                        pred='prob')
    print(perf)
    perf.to_excel(excel_file, sheet_name='模型评估')

    row_cnt = len(perf) + 2
    if sample_stat_group_fields is not None:
        for field in sample_stat_group_fields:
            perf = sample_df.groupby(field).apply(model_eval,
                                                  target=target,
                                                  pred='prob')
            print(perf)
            perf.to_excel(excel_file, sheet_name='模型评估', startrow=row_cnt)
            row_cnt += len(perf) + 2

    A = pdo / np.log(2)
    B = base_points - A * np.log(base_odds)

    sample_df['score'] = np.round(A * np.log(
        (1 - sample_df['prob']) / sample_df['prob']) + B)
    train_df = sample_df[(sample_df['_train_test_flag_'] == '01_train') &
                         sample_df[target].isin([0, 1])]
    _, breaks = gains_table(train_df[target],
                            train_df['score'],
                            return_breaks=True)
    breaks[0] = -np.inf
    breaks[-1] = np.inf
    train_gains_table = gains_table(train_df[target],
                                    train_df['score'],
                                    breaks=breaks)
    pd.DataFrame(['Train Gains Table']).to_excel(excel_file,
                                                 sheet_name='模型评估',
                                                 index=False,
                                                 header=False,
                                                 startrow=row_cnt)
    train_gains_table.to_excel(excel_file,
                               sheet_name='模型评估',
                               startrow=row_cnt + 1)
    row_cnt += len(train_gains_table) + 3

    if random_test_set:
        test_df = sample_df[(sample_df['_train_test_flag_'] == '02_test') &
                            sample_df[target].isin([0, 1])]
        test_gains_table = gains_table(test_df[target],
                                       test_df['score'],
                                       breaks=breaks)
        pd.DataFrame(['Test Gains Table']).to_excel(excel_file,
                                                    sheet_name='模型评估',
                                                    index=False,
                                                    header=False,
                                                    startrow=row_cnt)
        test_gains_table.to_excel(excel_file,
                                  sheet_name='模型评估',
                                  startrow=row_cnt + 1)
        row_cnt += len(train_gains_table) + 3

    oot_df = sample_df[(sample_df['_train_test_flag_'] == '03_oot') &
                       sample_df[target].isin([0, 1])]
    oot_gains_table = gains_table(oot_df[target],
                                  oot_df['score'],
                                  breaks=breaks)
    pd.DataFrame(['OOT Gains Table']).to_excel(excel_file,
                                               sheet_name='模型评估',
                                               index=False,
                                               header=False,
                                               startrow=row_cnt)
    oot_gains_table.to_excel(excel_file,
                             sheet_name='模型评估',
                             startrow=row_cnt + 1)
    row_cnt += len(oot_gains_table) + 3

    # 其他标签
    if alternative_target is not None:
        alternative_target = str_to_list(alternative_target)

        for y in alternative_target:
            perf = sample_df.groupby('_train_test_flag_').apply(model_eval,
                                                                target=y,
                                                                pred='prob')
            pd.DataFrame([y]).to_excel(excel_file,
                                       sheet_name='模型评估',
                                       index=False,
                                       header=False,
                                       startrow=row_cnt)
            perf.to_excel(excel_file, sheet_name='模型评估', startrow=row_cnt + 1)
            row_cnt += len(perf) + 3

    # PSI
    var_psi = woebin_psi(
        train_df,
        oot_df,
        bins={
            k: v for k, v in bins.items() if k + '_woe' in selected_variables
        })
    model_psi = pd.DataFrame({
        'variable': 'model_score',
        'bin': train_gains_table.index,
        'base_distr': train_gains_table['TotalPercent'],
        'cmp_distr': oot_gains_table['TotalPercent']
    }).assign(psi=lambda x: psi(x['base_distr'], x['cmp_distr']))

    psi_df = pd.concat([var_psi, model_psi], ignore_index=True)
    print(psi_df)
    psi_df.to_excel(excel_file, sheet_name='PSI分析', index=False)

    # Swap分析（如有）
    compare_model_fields = str_to_list(compare_model_fields)
    row_cnt = 0
    if compare_model_fields:
        for model in compare_model_fields:
            print(f'新模型 vs {model.upper()}\n')
            swap_tbl = swap_analysis_simple(sample_df,
                                            benchmark_model='score',
                                            challenger_model=model,
                                            target_col=target,
                                            reject_ratio=0.2)
            swap_tbl.to_excel(excel_file, sheet_name='Swap分析', startrow=row_cnt)
            row_cnt += len(swap_tbl) + 5
            print(swap_tbl, '\n\n')

    excel_file.close()

    woe_plots = woebin_plot({k[:-4]: bins[k[:-4]] for k in selected_variables})
    if not os.path.exists('pic'):
        os.mkdir('pic')

    if not os.path.isdir('pic'):
        raise NotADirectoryError('当前目录下pic路径不是目录，无法保存WOE图')
    else:
        for v, plt in woe_plots.items():
            plt.savefig(f'pic/{v}.png')



# -*- encoding: utf-8 -*-
"""
logging.py: 日志模块
"""

import logging as _logging
import sys
from logging import DEBUG, INFO, WARN, ERROR, FATAL

if sys.version_info[0] == 3 and sys.version_info[1] >= 8:
    pass
elif sys.version_info[0] == 3 and sys.version_info[1] in [6, 7]:
    import io
    import os
    import traceback

    def findCaller(self, stack_info=False, stacklevel=1):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = _logging.currentframe()
        #On some versions of IronPython, currentframe() returns None if
        #IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        orig_f = f
        while f and stacklevel > 1:
            f = f.f_back
            stacklevel -= 1
        if not f:
            f = orig_f
        rv = "(unknown file)", 0, "(unknown function)", None
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _logging._srcfile:
                f = f.f_back
                continue
            sinfo = None
            if stack_info:
                sio = io.StringIO()
                sio.write('Stack (most recent call last):\n')
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == '\n':
                    sinfo = sinfo[:-1]
                sio.close()
            rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
            break
        return rv

    def _log(self,
             level,
             msg,
             args,
             exc_info=None,
             extra=None,
             stack_info=False,
             stacklevel=1):
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.
        """
        sinfo = None
        if _logging._srcfile:
            #IronPython doesn't track Python frames, so findCaller raises an
            #exception on some versions of IronPython. We trap it here so that
            #IronPython can use logging.
            try:
                fn, lno, func, sinfo = self.findCaller(stack_info, stacklevel)
            except ValueError:  # pragma: no cover
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:  # pragma: no cover
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
        record = self.makeRecord(self.name, level, fn, lno, msg, args, exc_info,
                                 func, extra, sinfo)
        self.handle(record)

    _logging.Logger.findCaller = findCaller
    _logging.Logger._log = _log
else:
    raise RuntimeError(
        'Not support Python with version earlier than 3.6, get %s.%s' %
        (sys.version_info[0], sys.version_info[1]))

_logging.basicConfig(format='[%(asctime)s] %(levelname)s - '
                     '%(name)s.%(funcName)s(%(lineno)s): %(message)s',
                     level=ERROR)


def set_verbosity(level):
    logger = _logging.getLogger()
    logger.setLevel(level)


def _log(level, msg, *args, **kwargs):
    # DO NOT USE THIS FUNCTION DIRECTLY!
    module_name = sys._getframe(2).f_globals['__name__']
    logger = _logging.getLogger(module_name)
    logger.log(level, msg, stacklevel=3, *args, **kwargs)


def debug(msg, *args, **kwargs):
    _log(DEBUG, msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    _log(INFO, msg, *args, **kwargs)


def warn(msg, *args, **kwargs):
    _log(WARN, msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    _log(ERROR, msg, *args, **kwargs)


def fatal(msg, *args, **kwargs):
    _log(FATAL, msg, *args, **kwargs)


__all__ = [
    'DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL', 'debug', 'info', 'warn', 'error',
    'fatal', 'set_verbosity'
]



# -*- encoding: utf-8 -*-
from typing import Tuple, Union, Dict

import numpy as np
import pandas as pd
from scipy.stats import ks_2samp
from sklearn.metrics import roc_auc_score

from syriskmodels.utils import sample_stats


def ks_score(y_true, y_pred) -> float:
    """
    计算模型 KS 统计量

    Args
        - y_true: 真实Y标，0-1变量
        - y_pred: 预测值

    Returns
        KS值
    """
    return ks_2samp(y_pred[y_true == 1], y_pred[y_true == 0])[0]


def model_perf(y_true, y_pred) -> Dict[str, float]:
    """
    评估模型表现，该函数封装了`sklearn.metrics.roc_auc_score` 和 `ks_score` 两个函数
    Args:
        y_true:
        y_pred:

    Returns:
        `{'auc': auc score, 'ks': ks score}`

    """
    labeled_selector = (np.isin(y_true, [0, 1])) & (~y_pred.isna())
    y_true_new = y_true[labeled_selector]
    y_pred_new = y_pred[labeled_selector]

    if np.sum(y_true_new) > 0 and np.prod(y_true_new) == 0:
        auc = roc_auc_score(y_true_new, y_pred_new)
        if auc < 0.5:
            auc = 1 - auc
        ks = ks_score(y_true_new, y_pred_new)
    else:
        auc = ks = np.nan

    return {'auc': auc, 'ks': ks}


def model_eval(df, target, pred) -> pd.Series:
    perf = model_perf(df[target], df[pred])
    auc = perf['auc']
    ks = perf['ks']
    tmp_df = df[df[target].isin([0, 1])]
    bad_rate = np.mean(tmp_df[target])
    return pd.Series([bad_rate, auc, ks], index=['bad_rate', 'auc', 'ks'])


def gains_table(
    y_true,
    y_score,
    split=None,
    breaks=None,
    ascending=True,
    return_breaks=False
) -> Union[Tuple[pd.DataFrame, np.ndarray], pd.DataFrame]:
    """
    A gains table includes a distribution of total, good, and bad cases by
    individual scores or score ranges.

    Args:
        y_true: 样本真实标签
        y_score: 样本对应模型分数
        split: 等分段数上限，默认 10。使用该参数时，不可同时指定 breaks 参数
        breaks: 分段点，默认为空。使用该参数时，不可同时指定 split 参数
        ascending: 为 True 时 gains table 按照分数由低到高统计；否则分数由高到低。
        return_breaks: 是否返回 breaks 分段点

    Returns:
        当 return_breaks=False 时，返回 gains_table (pd.DataFrame)；否则，返回
        元组 (gains_table, breaks)

    """
    if split is not None and breaks is not None:
        assert (
            'Either split or breaks should be None. Got split=%s, breaks=%s' %
            (split, breaks))
    if split is None and breaks is None:
        split = 10

    pred_df = pd.DataFrame({'score': y_score, 'y': y_true})
    if breaks is None:
        _, breaks = pd.qcut(
            pred_df.score,
            np.linspace(0, 1, split + 1),
            retbins=True,
            duplicates='drop')
    else:
        breaks = np.asarray(breaks)

    pred_df['score_seg'] = pd.cut(pred_df.score, bins=breaks)
    gains_df = pred_df.groupby('score_seg').apply(
        sample_stats, target='y').sort_index(ascending=ascending)
    gains_df = gains_df.assign(
        Odds=lambda x: x.GoodCnt / x.BadCnt,
        CumBadRate=lambda x: x.BadCnt.cumsum() / x.TotalCnt.cumsum(),
        BadPercent=lambda x: x.BadCnt / x.BadCnt.sum(),
        CumBadPercent=lambda x: x.BadCnt.cumsum() / x.BadCnt.sum(),
        GoodPercent=lambda x: x.GoodCnt / x.GoodCnt.sum(),
        CumGoodPercent=lambda x: x.GoodCnt.cumsum() / x.GoodCnt.sum(),
        TotalPercent=lambda x: x.TotalCnt / x.TotalCnt.sum(),
        Lift=lambda x: (x.BadCnt / x.TotalCnt) /
        (x.BadCnt.sum() / x.TotalCnt.sum())).assign(
            KS=lambda x: np.abs(x.CumBadPercent - x.CumGoodPercent))

    gains_df = gains_df[[
        'TotalCnt', 'GoodCnt', 'BadCnt', 'Odds', 'BadRate', 'Lift',
        'CumBadRate', 'BadPercent', 'CumBadPercent', 'GoodPercent',
        'CumGoodPercent', 'KS', 'TotalPercent'
    ]]

    if return_breaks:
        return gains_df, breaks
    else:
        return gains_df


def psi(base_distr, cmp_distr, epsilon=1e-3):
    """计算PSI (Population Stability Index)"""

    def distr_preprocess(distr_arr):
        distr_arr = np.asarray(distr_arr)
        distr_arr = np.where(np.isnan(distr_arr), 0, distr_arr)
        distr_arr = distr_arr / np.sum(distr_arr)

        # 填充0值
        if np.any(distr_arr == 0):
            distr_arr = distr_arr + epsilon
            distr_arr = distr_arr / base_distr.sum()

        return distr_arr

    base_distr = distr_preprocess(base_distr)
    cmp_distr = distr_preprocess(cmp_distr)

    # calculate psi
    psi_value = (base_distr - cmp_distr) * np.log(base_distr / cmp_distr)
    return psi_value.sum()


def swap_analysis(df,
                  benchmark_model,
                  challenger_model,
                  target_col,
                  benchmark_breaks=None,
                  challenger_breaks=None,
                  segments=10,
                  right=False,
                  retval='badrate'):
    """
    双模型swap分析
    Args:
        df: 包含模型分及Y标的数据框
        benchmark_model: 基准模型
        challenger_model: 挑战者模型
        target_col: Y标列名
        benchmark_breaks: 基准模型切分点，默认为空
        challenger_breaks: 挑战者模型切分点，默认为空
        segments: 等分段数，默认为10段
        right: right=True时为左开右闭区间，否则为右闭左开区间
        retval: 返回统计量，
            'badrate'返回坏率，
            'badcnt'返回坏样本数，
            'totalcnt'返回总样本数，
            'all'返回'totalcnt', 'badcnt', 'badrate'构成的元组

    Returns:
        见 retval 参数说明

    """
    assert retval in ['badrate', 'badcnt', 'totalcnt', 'all'], ValueError(
        'argument retval should be one of \'badrate\', \'badcnt\', '
        f'\'totalcnt\', and \'all\', get {retval}')
    df = df[df[target_col].isin([0, 1]) & (~df[benchmark_model].isna()) &
            (~df[challenger_model].isna())].copy()

    if benchmark_breaks is None:
        df[benchmark_model + '_seg'] = pd.qcut(
            df[benchmark_model], np.linspace(0, 1, segments + 1), duplicates='drop')
    else:
        df[benchmark_model + '_seg'] = pd.cut(
            df[benchmark_model], benchmark_breaks, right=right)

    if challenger_breaks is None:
        df[challenger_model + '_seg'] = pd.qcut(
            df[challenger_model],
            np.linspace(0, 1, segments + 1),
            duplicates='drop')
    else:
        df[challenger_model + '_seg'] = pd.cut(
            df[challenger_model], challenger_breaks, right=right)
    total_cnt = pd.pivot_table(
        df,
        index=benchmark_model + '_seg',
        columns=challenger_model + '_seg',
        values=target_col,
        aggfunc='count')
    if retval == 'totalcnt':
        return total_cnt

    bad_cnt = pd.pivot_table(
        df,
        index=benchmark_model + '_seg',
        columns=challenger_model + '_seg',
        values=target_col,
        aggfunc='sum')
    if retval == 'badcnt':
        return bad_cnt

    bad_rate = bad_cnt / total_cnt
    if retval == 'badrate':
        return bad_rate

    return total_cnt, bad_cnt, bad_rate


def swap_analysis_simple(df,
                         benchmark_model,
                         challenger_model,
                         target_col,
                         reject_ratio=0.2):
    """
    简易双模型swap分析，该函数将两个模型等量的尾部客户标记为拒绝样本，尾部客户占比由`reject_ratio`参数指定，并
    计算换入换出的样本坏率。**注意：该函数要求模型分越高，风险越低**
    Args:
        df: 包含模型分及Y标的数据框
        benchmark_model: 基准模型
        challenger_model: 挑战者模型
        target_col: Y标列名
        reject_ratio: 待拒绝尾部客户占比

    Returns:
        `pd.DataFrame`
    """
    df = df[df[target_col].isin([0, 1]) & (~df[benchmark_model].isna()) &
            (~df[challenger_model].isna())].copy()
    df[benchmark_model + '_seg2'] = pd.qcut(
        df[benchmark_model], [0, reject_ratio, 1.], duplicates='drop')
    df[challenger_model + '_seg2'] = pd.qcut(
        df[challenger_model], [0, reject_ratio, 1.], duplicates='drop')
    total_cnt2 = pd.pivot_table(
        df,
        index=benchmark_model + '_seg2',
        columns=challenger_model + '_seg2',
        values=target_col,
        aggfunc='count')
    bad_cnt2 = pd.pivot_table(
        df,
        index=benchmark_model + '_seg2',
        columns=challenger_model + '_seg2',
        values=target_col,
        aggfunc='sum')
    bad_rate2 = bad_cnt2 / total_cnt2
    swap_result = pd.concat([total_cnt2 / np.sum(total_cnt2.values), bad_rate2],
                            axis=1)
    swap_result.columns = [['total%', 'total%', 'bad_rate', 'bad_rate'],
                           swap_result.columns.tolist()]
    swap_result.index.name = 'base'
    swap_result.columns.name = 'cmp'

    return swap_result



# -*- encoding: utf-8 -*-
import numpy as np
import pandas as pd
import unittest

from addpath import addpath

try:
    addpath()
    from syriskmodels.detector import detect
except ModuleNotFoundError as e:
    raise e


class TestDetect(unittest.TestCase):

    def test_detect(self):
        arr = np.random.rand(50, 100)
        df = pd.DataFrame(arr, columns=['V' + str(i) for i in range(100)])
        res1 = detect(df, n_cores=1)
        res2 = detect(df, n_cores=None)
        res2 = detect(df, n_cores=None)


if __name__ == '__main__':
    unittest.main()


# -*- encoding: utf-8 -*-
from types import UnionType
from typing import List, Union

import numpy as np
import pandas as pd

from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import check_cv

import syriskmodels.logging as logging


class LogisticRegressionCV(object):

  def __init__(self, *, cv=3, **logistic_kwargs):
    self._cv = cv
    self._logistic_kwargs = logistic_kwargs

  def fit_and_eval(self, X, y):
    train_predicted = None
    train_actual = None
    valid_predicted = None
    valid_actual = None

    cv_wrapper = check_cv(self._cv, y=y, classifier=True)

    for train_index, test_index in cv_wrapper.split(X, y):
      X_train, X_valid = X[train_index], X[test_index]
      y_train, y_valid = y[train_index], y[test_index]

      logistic_regressor = LogisticRegression(**self._logistic_kwargs)
      logistic_regressor.fit(X_train, y_train)

      train_pred = logistic_regressor.predict_proba(X_train)[:, 1]
      if train_predicted is None:
        train_predicted = train_pred
        train_actual = y_train
      else:
        train_predicted = np.concatenate([train_predicted, train_pred])
        train_actual = np.concatenate([train_actual, y_train])

      valid_pred = logistic_regressor.predict_proba(X_valid)[:, 1]
      if valid_predicted is None:
        valid_predicted = valid_pred
        valid_actual = y_valid
      else:
        valid_predicted = np.concatenate([valid_predicted, valid_pred])
        valid_actual = np.concatenate([valid_actual, y_valid])

    train_auc = roc_auc_score(train_actual, train_predicted)
    valid_auc = roc_auc_score(valid_actual, valid_predicted)
    logging.debug(f'训练集 AUC={train_auc:.2%}，测试集 AUC={valid_auc:.2%}')

    return min(train_auc, valid_auc) - abs(train_auc - valid_auc)


def group_split_cv(group_arr):
  group_arr = np.asarray(group_arr)
  groups = np.unique(group_arr)

  for g in groups:
    train = group_arr != g
    yield np.argwhere(train).ravel(), np.argwhere(~train).ravel()


def stepwise_lr(df: pd.DataFrame,
                y: str,
                x: Union[str, List[str]],
                cv: int = 3,
                max_num_features: int = 30,
                initial_features: Union[str, List[str]] = None,
                direction: str = 'bidirectional',
                **lr_kwargs):
  feature_pool = x
  if initial_features is None:
    selected_features = []
  else:
    if isinstance(initial_features, str):
      selected_features = [initial_features]
    elif not isinstance(initial_features, list):
      selected_features = list(initial_features)
    else:
      selected_features = initial_features
  best_metrics = None
  assert direction in ('forward', 'backward', 'bidirectional'), \
    f'direction参数为 forward, backward, bidirectional 三者之一，输入{direction}不合法'

  def get_features_perf(feature_list):
    lr_cv = LogisticRegressionCV(cv=cv, **lr_kwargs)
    auc = lr_cv.fit_and_eval(df[feature_list].values, df[y])
    return auc

  step = 0
  while True:
    perf_records = {}
    improved = False
    step += 1

    if direction in ['forward', 'bidirectional']:
      for feature in feature_pool:
        train_features = selected_features + [feature]
        perf_records[tuple(train_features)] = get_features_perf(train_features)

      if len(selected_features) > 1:
        for feature in selected_features:
          train_features = selected_features.copy()
          train_features.remove(feature)
          perf_records[tuple(train_features)] = get_features_perf(train_features)

    if direction in ['backward', 'bidirectional']:
      for key, value in perf_records.items():
        if best_metrics is None or (best_metrics < value and
                                    len(key) <= max_num_features):
          selected_features = list(key)
          best_metrics = value
          improved = True
          feature_pool = [f for f in x if f not in selected_features]

    if improved:
      logging.info(f'Step {step}:\nSelected features: {selected_features}\n'
                   f'Performance: auc={best_metrics}')
    else:
      logging.info('No improve. Exit.')
      break

  return best_metrics, selected_features



# -*- encoding: utf-8 -*-
__version__ = '0.2.2'

import syriskmodels.contrib as contrib
import syriskmodels.detector as detector
import syriskmodels.evaluate as evaluate
import syriskmodels.logging as logging
import syriskmodels.models as models
import syriskmodels.scorecard as scorecard
import syriskmodels.utils as utils
from .detector import detect
from .evaluate import gains_table
from .evaluate import ks_score
from .evaluate import model_eval
from .models import stepwise_lr
from .scorecard import make_scorecard
from .scorecard import woebin
from .scorecard import woebin_breaks
from .scorecard import woebin_plot
from .scorecard import woebin_ply
from .scorecard import woebin_psi
from .utils import monotonic
from .utils import sample_stats

__all__ = [
    'contrib', 'detector', 'evaluate', 'logging', 'models', 'scorecard',
    'utils', 'detect', 'gains_table', 'ks_score', 'model_eval', 'stepwise_lr',
    'make_scorecard', 'woebin', 'woebin_breaks', 'woebin_plot', 'woebin_ply',
    'woebin_psi', 'monotonic', 'sample_stats'
]



# -*- encoding: utf-8 -*-

import pandas as pd

from syriskmodels.scorecard import woebin
from syriskmodels.scorecard import woebin_breaks


def risk_trends_consistency(df, sc_bins, target):
    """
    判断变量在不同数据集上风险趋势与分箱结果中风险趋势是否一致

    Args:
        df: pd.DataFrame
        sc_bins: syriskmodels.scorecard.woebin 返回的结果
        target: 目标变量的列名

    Returns:
        {变量名: 分箱badprob序列的 Spearman 秩相关系数}，秩相关系数有效区间 [-1,1]；
        其中 -1 代表趋势完全相反，1 代表趋势完全一致。

    """
    bin_var = list(sc_bins.keys())
    df_var = df.columns.tolist()
    variables = list(set(bin_var).intersection(df_var))

    brk, spc_val = woebin_breaks(sc_bins)
    df_bins = woebin(df,
                     x=variables,
                     y=target,
                     breaks_list=brk,
                     special_values=spc_val)

    consistency = {}
    for v in variables:
        old_bin = sc_bins[v]
        old_bin = old_bin[~old_bin['is_special_values']]
        new_bin = df_bins[v]
        bin_merge = pd.merge(old_bin[['variable', 'bin', 'badprob']],
                             new_bin[['variable', 'bin', 'badprob']],
                             on=['variable', 'bin'],
                             suffixes=('_o', '_n'),
                             how='left')
        c = bin_merge[['badprob_n',
                       'badprob_o']].corr(method='spearman').iloc[0, 0]
        consistency[v] = c

    return consistency



# -*- encoding: utf-8 -*-
"""
scorecard.py - 变量分箱算法

本模块从 scorecardpy(https://github.com/ShichenXie/scorecardpy)移植，并对代码进行了
大量重构，以提供更好的扩展性和向后兼容性。本模块主要接口尽可能保持与 scorecardpy 相同，主要有
如下接口函数：

* woebin
* woebin_ply
* woebin_plot

此外，提供以下接口函数：

* sc_bins_to_df
* make_scorecard
* woebin_breaks
* woebin_psi
"""
import itertools
import multiprocessing as mp
import re
import time
from typing import Optional
from typing import Union

import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype
from scipy.stats import chi2
from scipy.stats import chi2_contingency
from scipy.stats import fisher_exact

import syriskmodels.logging as logging
from syriskmodels.evaluate import psi
from syriskmodels.utils import monotonic
from syriskmodels.utils import round_
from syriskmodels.utils import str_to_list

OK = 0
CONST = 10
TOO_MANY_VALUES = 20


def check_uniques(s: pd.Series, max_cate_num: int = 50):
  """
  检查变量非空单一值数，当非空单一值数小于等于1时，该变量为常量，不适合用于建模；当变量为类别变量（非数值型）时，
  过多类别值也不适合用于建模。对满足上述情况的变量，本函数返回 False；否则返回 True。
  Args:
    s: pd.Series
    max_cate_num: int, 最大允许类别数

  Returns: bool

  """
  n_uniques = len(np.unique(s[~s.isna()]))
  if n_uniques <= 1:
    return CONST
  elif (not is_numeric_dtype(s)) and n_uniques > max_cate_num:
    return TOO_MANY_VALUES
  else:
    return OK


def replace_blank_string(s: pd.Series):
  return s.replace('', np.nan)


def check_y(dat: pd.DataFrame, y: str, *, positive: Union[int, float] = 1):
  """
  对dat中y标进行如下检查：(1)dat[y]是否为数值型；(2)dat[y]空值检查；(3)dat[y]是否为(0,1)二值型变量.
  完成检查后，将dat[y]等于positive的记录记为1（正样本），否则记为0.
  """
  dat = dat.copy()
  try:
    if not is_numeric_dtype(dat[y]):
      msg = f'The target column {y} is not numeric type, dtype={dat[y].dtypes}'
      logging.error(msg)
      raise TypeError(msg)

    if dat[y].isna().any():
      logging.warn(f'There are NaNs in {y} column. '
                   f'The corresponding records are removed.')
      dat = dat.dropna(subset=[y])

    if dat[y].nunique() != 2:
      logging.error(f'The target column {y} is not binary.')
      raise ValueError(f'The target column {y} is not binary.')

    if np.all(dat[y] != positive):
      logging.error(f'positive value ({positive}) not in the target column.')
      raise ValueError(f'positive value ({positive}) not in the target column.')

    dat[y] = np.where(dat[y] == positive, 1, 0)

  except KeyError as err:
    logging.error(
        "Incorrect inputs; there is no \'{}\' column in dat.".format(y))
    raise KeyError(y) from err

  return dat


def x_variable(dat, y, x, var_skip=None):
  y = str_to_list(y)
  if var_skip is not None:
    y = y + str_to_list(var_skip)
  x_all = list(set(dat.columns) - set(y))

  if x is None:
    x = x_all
  else:
    x = str_to_list(x)

    if any([i in list(x_all) for i in x]) is False:
      x = x_all
    else:
      x_not_in_x_all = set(x).difference(x_all)
      if len(x_not_in_x_all) > 0:
        logging.warn("Incorrect inputs; there are {} x variables are not exist "
                     "in input data, which are removed from x. \n({})".format(
                         len(x_not_in_x_all), ', '.join(x_not_in_x_all)))
        x = set(x).intersection(x_all)

  return list(x)


def check_breaks_list(breaks_list) -> dict:
  if breaks_list is not None:
    # is string
    if isinstance(breaks_list, str):
      breaks_list = eval(breaks_list)
    # is not dict
    if not isinstance(breaks_list, dict):
      raise Exception("Incorrect inputs; breaks_list should be a dict.")
  else:
    breaks_list = {}
  return breaks_list


def check_special_values(special_values, xs) -> dict:
  if special_values is not None:
    # # is string
    # if isinstance(special_values, str):
    #     special_values = eval(special_values)
    if isinstance(special_values, list):
      logging.warn(
          "The special_values should be a dict. Make sure special values "
          "are exactly the same in all variables if special_values is "
          "a list.")
      sv_dict = {}
      for i in xs:
        sv_dict[i] = special_values
      special_values = sv_dict
    elif not isinstance(special_values, dict):
      raise Exception(
          "Incorrect inputs; special_values should be a list or dict.")
  else:
    special_values = {}
  return special_values


def woebin(
    dt,
    y,
    x=None,
    var_skip=None,
    breaks_list=None,
    special_values=None,
    # positive="bad|1",
    positive=1,
    no_cores=None,
    methods=None,
    max_cate_num=50,
    replace_blank=np.nan,
    **kwargs):
  if methods is None:
    methods = ['quantile', 'tree']

  # start time
  start_time = time.time()
  # arguments
  dt = dt.copy(deep=True)
  y = str_to_list(y)
  x = str_to_list(x)
  if x is not None:
    dt = dt[y + x]
  # check y
  dt = check_y(dt, y[0], positive=positive)

  # x variable names
  xs = x_variable(dt, y, x, var_skip)
  xs_len = len(xs)
  # breaks_list
  breaks_list = check_breaks_list(breaks_list)
  # special_values
  special_values = check_special_values(special_values, xs)

  # binning for each x variable
  # loop on xs
  if (no_cores is None) or (no_cores < 1):
    all_cores = mp.cpu_count() - 1
    no_cores = int(
        np.ceil(xs_len / 5 if xs_len / 5 < all_cores else all_cores * 0.9))

  # y list to str
  y = y[0]

  woe_bin = WOEBinFactory.build(methods, **kwargs)

  tasks = [
      (
          # dtm definition
          pd.DataFrame({
              'y': dt[y],
              'variable': x_i,
              'value': dt[x_i]
          }),
          # breaks_list
          breaks_list.get(x_i),
          # special_values
          special_values.get(x_i),
          max_cate_num,
          replace_blank,
      ) for x_i in xs
  ]

  logging.info(f'开始分箱，特征数 {len(tasks)}，样本数 {len(dt)}')

  if no_cores == 1:
    bins = dict(zip(xs, itertools.starmap(woe_bin, tasks)))
  else:
    pool = mp.Pool(processes=no_cores)
    bins = dict(zip(xs, pool.starmap(woe_bin, tasks)))
    pool.close()

  # running time
  running_time = time.time() - start_time
  logging.info('分箱完成：Binning on {} rows and {} columns in {}'.format(
      dt.shape[0], len(xs), time.strftime("%H:%M:%S",
                                          time.gmtime(running_time))))

  return bins


class WOEBinFactory(object):
  """WOEBin工厂方法类，与ComposedWOEBin配合，提供WOEBin子类的组装功能。

    本模块原生提供了两种细分箱方法：QuantileInitBin与HistogramInitBin，两种粗分箱方法：
    ChiMergeOptimBin与TreeOptimBin。一般在实际使用时需要将细分箱与粗分箱结合使用，例如：
    先用QuantileInitBin进行等频分箱，然后再用TreeOptimBin对进行细分箱。

    实现上述功能的代码如下：

    >>> woebin = WOEBinFactory.build(['quantile', 'tree'])
    >>> woebin(dtm)

    其中，quantile和tree为上述两个类的注册名，也可显示写出类名如下：

    >>> woebin = WOEBinFactory.build([QuantileInitBin, TreeOptimBin])
    >>> woebin(dtm)

    详情请见`WOEBinFactory.register`和`WOEBinFactory.build`方法文档。
    """
  __woebin_class_mapping = {}

  @classmethod
  def register(cls, names):
    """
        注册分箱类的装饰器。对分箱类使用该装饰器并指定注册名称后，在`build`方法中就可以使用
        注册名称替代类名。使用方法可参考`QuantileInitBin`, `HistogramInitBin`,
        `TreeOptimBin`以及`ChiMergeOptimBin`四个分箱类。

        Args:
            names: str or list[str]，分箱类的注册名称

        Returns:
            该装饰器装饰的WOEBin子类

        """
    names = str_to_list(names)

    def wrapped(bin_class):
      if not issubclass(bin_class, WOEBin):
        raise TypeError(f'类{bin_class}不是WOEBin子类，无法注册')

      for name in names:
        if name in cls.__woebin_class_mapping.keys():
          raise KeyError(f'名称{name}已存在，'
                         f'类{bin_class.__name__}不能注册为{name}')
        else:
          cls.__woebin_class_mapping[name] = bin_class

      return bin_class

    return wrapped

  @classmethod
  def get_binner(cls, bin_class, **kwargs):
    if isinstance(bin_class, str):
      try:
        bin_class = cls.__woebin_class_mapping[bin_class]
      except KeyError:
        raise KeyError(f'方法{bin_class}未注册！')

    # if issubclass(bin_class, WOEBin):
    #     binner = bin_class(**kwargs)
    # elif isinstance(bin_class, WOEBin):
    #     binner = bin_class
    if isinstance(bin_class, WOEBin):
      binner = bin_class
    elif isinstance(bin_class, type) and issubclass(bin_class, WOEBin):
      binner = bin_class(**kwargs)
    else:
      raise TypeError(f'类{bin_class}不是WOEBin实例或子类')

    return binner

  @classmethod
  def build(cls, bin_classes, **kwargs):
    """
        将多个WOEBin*按顺序*组装为一个WOEBin子类（ComposedWOEBin）,该方法可接受任意关键字
        参数，并传递给子类初始化方法。通常的组装方式为细分箱类(*InitBin) +
        粗分箱类(*OptimBin)。

        >>> woe_bin = WOEBinFactory.build (['quantile', 'tree'],
        ...                               initial_bins=20,
        ...                               bin_num_limit=8,
        ...                               min_iv_inc=0.1,
        ...                               count_distr_limit=0.05)
        >>> woe_bin
        <syriskmodels.scorecard.ComposedWOEBin object at 0x1009776a0>

        Args:
            bin_classes: WOEBin子类或子类注册名列表
            **kwargs: 子类初始化的关键字参数

        Returns:
            ComposedWOEBin实例

        """
    bin_objects = [cls.get_binner(bin_cls, **kwargs) for bin_cls in bin_classes]
    return ComposedWOEBin(bin_objects, **kwargs)


class WOEBin(object):
  """WOEBin: 对单个变量进行分箱操作的基类，所有分箱操作的类都继承本类。

    分箱类型包括细分箱(*InitBin)、粗分箱(*OptimBin)。其中细分箱用来初始化分箱，一般是通
    过等频(quantile)或等宽(histogram)的方式进行分箱；粗分箱在粗分箱结果的基础上进行，用
    来简化分箱结果，确保分箱结果的显著性和稳定性。

    WOEBin实例为可调用(callable)对象，接受dtm、breaks、special_values三个参数，当
    breaks 不为空时，WOEBin实例将直接使用breaks提供的切分点进行分箱，不再调用`woebin`方法；
    当 special_values 不为空时，调用WOEBin实例会把原数据集拆分为特殊值、非特殊值两部分，
    对特殊值部分，每个特殊值将单独成箱，不做分箱合并，对于非特殊值按照breaks参数或调用woebin
    方法返回的切分点进行分箱。注意：当原数据集中包含空值时，空值将作为特殊值单独成箱。

    `woebin`方法为分箱主方法，接收数据集中非特殊值（排除空值和special_values中指定的其他
    特殊值）部分作为入参，并返回*切分点*。扩展本类时，需要重写本方法。对于细分箱(*InitBin)，
    该方法不接受breaks参数；对于粗分箱(*OptimBin)，需要为本方法传入细分箱`woebin`方法返回
    的切分点作为breaks入参。

    分箱名约定：当x为数值型变量时，分箱名为 [a,b) 左闭右开区间；当x为类别变量时，分箱名为
    c1%,%c2%,%...%,%cn，即用%,%拼接的类别名。

    切分点约定：当x为数值型变量，分箱名为 [a,b) 时，切分点为区间右边界 b；当x为类别变量时，
    切分点与分享名相同。

    * 常用变量约定

    dtm定义:
        下述所有dtm变量为（至少）包含 variable（变量名）、y（目标变量）、value（解释
        变量值）三列的数据框（pd.DataFrame）。其中value可包含空值和其他特殊值

    ns_dtm：不含空值、特殊值的dtm（ns for non-special）

    binning_*：
        分箱信息统计表，包含 variable（变量名）、bin_chr（分箱名）、good（好样本数）、
        bad（坏样本数）四列的数据框（pd.DataFrame）。

    Args
        eps: 若某一分箱中好样本或坏样本数为0，计算woe及iv时用eps替换0，默认值0.5
    """

  def __init__(self, eps=0.5, **kwargs):
    self.epsilon = eps
    self.kwargs = kwargs

  @staticmethod
  def add_missing_spl_val(dtm: pd.DataFrame, spl_val):
    """如果数据集中存在空值，则将 'missing' 加入到special_values中"""
    special_values = spl_val
    if dtm['value'].isnull().any():
      if spl_val is None:
        special_values = ['missing']
      elif 'missing' not in spl_val:
        special_values = ['missing'] + spl_val

    return special_values

  @staticmethod
  def split_vec_to_df(vec):
    """
        特殊值/断点列表转DataFrame, 如下例所示：

        >>> vec
        ['missing', '1', '2', '3%,%4']
        >>> WOEBin.split_vec_to_df(vec)
           bin_chr  rowid value
        0  missing      0   NaN
        1        1      1     1
        2        2      2     2
        3    3%,%4      3     3
        4    3%,%4      3     4

        Args:
            vec: 特殊值列表

        Returns:
            pd.DataFrame，包含如下四列 'bin_char', 'rowid', 'value'

        """
    assert vec is not None, 'vec cannot be None'
    vec = [str(i) for i in vec]
    a = pd.DataFrame({'bin_chr': vec}).assign(rowid=lambda x: x.index)
    b = pd.DataFrame([i.split('%,%') for i in vec], index=vec).stack().replace(
        'missing', np.nan).reset_index(name='value').rename(
            columns={'level_0': 'bin_chr'})[['bin_chr', 'value']]

    df = pd.merge(a, b, on='bin_chr')
    return df

  @classmethod
  def split_special_values(cls, dtm, spl_val):
    dtm['idx'] = dtm.index
    spl_val = cls.add_missing_spl_val(dtm, spl_val)
    if spl_val is not None:
      sv_df = cls.split_vec_to_df(spl_val)
      # value
      if is_numeric_dtype(dtm['value']):
        sv_df['value'] = sv_df['value'].astype(dtm['value'].dtypes)
        # TODO: 此处是否必要？？
        sv_df['bin_chr'] = np.where(
            np.isnan(sv_df['value']), sv_df['bin_chr'],
            sv_df['value'].astype(str))
      # dtm_sv & dtm
      dtm_merge = pd.merge(
          dtm.fillna("missing"),
          sv_df[['value', 'rowid']].fillna("missing"),
          how='left',
          on='value')
      dtm_sv = dtm_merge[~dtm_merge['rowid'].isna()][
          dtm.columns.tolist()].reset_index(drop=True)
      dtm_ns = dtm_merge[dtm_merge['rowid'].isna()][
          dtm.columns.tolist()].reset_index(drop=True)
      if len(dtm_ns) == 0:
        dtm_ns = None
      else:
        dtm_ns['value'] = dtm_ns['value'].astype(dtm['value'].dtypes)

      if dtm_sv.shape[0] == 0:
        dtm_sv = None
      else:
        dtm_sv = pd.merge(
            dtm_sv.fillna('missing'), sv_df.fillna('missing'), on='value')
    else:
      dtm_sv = None
      dtm_ns = dtm.copy()

    if dtm_sv is not None:
      dtm_sv.set_index(dtm_sv['idx'], drop=True, inplace=True)

    if dtm_ns is not None:
      dtm_ns.set_index(dtm_ns['idx'], drop=True, inplace=True)

    return {'dtm_sv': dtm_sv, 'dtm_ns': dtm_ns}

  @classmethod
  def dtm_binning_sv(cls, dtm, spl_val):
    """
    将原数据集拆分为特殊值数据集、非特殊值数据集，特殊值列表有spl_val参数给出。

    Specifications:
    1、当dtm['value']为数值类型时，每个特殊数值会单独成为一箱。例如：
    >>> vec = ['missing', '-9999%,%-999']
    如果dtm['value']为数值类型，在此函数中，-9999和-999会分开成为两箱；如果dtm['value']为
    非数值类型，则-9999和-999会合并到同一分箱。
    2、如果spl_val为None，或者dtm['value']中不存在特殊值，则返回的binning_sv为None。
    3、如果dtm['value']全为spl_val中的特殊值，则返回的dtm为None。
    4、如果入参breaks中包含'missing'，则spl_val中将不包含'missing'。

    Args:
        dtm: 应包含['y', 'variable', 'value']三列的DataFrame，详见WEOBin类文档
        spl_val: 特殊值列表

    Returns:
        {'binning_sv': 特殊值分箱统计结果, 'ns_dtm': dtm中除去特殊值外的部分}

    """
    split_dtm = cls.split_special_values(dtm, spl_val)

    dtm_sv = split_dtm['dtm_sv']
    dtm_ns = split_dtm['dtm_ns']

    if dtm_sv is None or dtm_sv.shape[0] == 0:
      binning_sv = None
    else:
      binning_sv = cls.binning(dtm_sv, dtm_sv['bin_chr'])

    return {'binning_sv': binning_sv, 'ns_dtm': dtm_ns}

  def __call__(self,
               dtm,
               breaks=None,
               special_values=None,
               max_cate_num=50,
               replace_blank=np.nan):
    dtm['value'] = replace_blank_string(dtm['value'])
    ret = check_uniques(dtm['value'], max_cate_num)
    if ret == CONST:
      return 'CONST'
    elif ret == TOO_MANY_VALUES:
      return 'TOO_MANY_VALUES'
    else:
      binning_split = self.dtm_binning_sv(dtm, special_values)
      binning_sv = binning_split['binning_sv']
      dtm = binning_split['ns_dtm']

      if dtm is None:
        binning_dtm = None
      else:
        if breaks is not None:
          binning_dtm = self.binning_breaks(dtm, breaks)
        else:
          breaks = self.woebin(dtm)
          binning_dtm = self.binning_breaks(dtm, breaks)

      bin_list = {'binning_sv': binning_sv, 'binning': binning_dtm}
      binning = pd.concat(bin_list, keys=bin_list.keys())
      binning = binning.reset_index()
      binning = binning.assign(is_sv=lambda x: x.level_0 == 'binning_sv')

      return self.binning_format(binning)

  def woebin(self, dtm, breaks=None):
    """"""
    raise NotImplementedError

  @classmethod
  def binning(cls, dtm, bin_chr):
    """
    给定dtm、分箱名序列生成binning，dtm和binning的定义见WOEBin类文档。bin_chr为
    pd.Series，长度与dtm相同，对应dtm中每个样本所属的分箱名称。

    Args:
      dtm: dtm定义见`WOEBin`类文档
      bin_chr: dtm中每个样本对应的分箱名称，pd.Series。

    Returns: binning，定义见`WOEBin`类文档
    """

    def _n0(x):
      return np.sum(x == 0)

    def _n1(x):
      return np.sum(x == 1)

    bin_chr = bin_chr.rename(index='bin_chr')
    binning = dtm.groupby(['variable', bin_chr], observed=False)['y'].agg(
        good=_n0, bad=_n1)
    binning = binning.reset_index()

    return binning

  @classmethod
  def apply(cls, dtm, bin_res, value='woe'):
    """
    Args:
      dtm:
      bin_res:
      value: 转换值类型，可选项('woe', 'index', 'bin')，默认woe.

    Returns:

    """
    special_values = bin_res['breaks'][bin_res['is_special_values']].tolist()
    if 'missing' in special_values:
      special_values.remove('missing')
    if len(special_values) == 0:
      special_values = None
    breaks = bin_res['breaks'][~bin_res['is_special_values']]
    dtm['value'] = replace_blank_string(dtm['value'])
    split_dtm = cls.split_special_values(dtm, special_values)
    dtm_sv = split_dtm['dtm_sv']
    dtm_ns = split_dtm['dtm_ns']

    if dtm_sv is not None:
      dtm_sv = dtm_sv[['idx', 'bin_chr', 'value', 'y']]

    if dtm_ns is not None:
      break_df = cls.split_vec_to_df(breaks)
      # TO-DO: 以下代码与binning_breaks重复，需要进行精简
      if is_numeric_dtype(dtm_ns['value']):
        break_list = ['-inf'] + list(
            set(break_df.value.tolist()).difference(
                {np.nan, '-inf', 'inf', 'Inf', '-Inf'})) + ['inf']
        break_list = sorted(list(map(float, break_list)))
        labels = [
            '[{},{})'.format(break_list[i], break_list[i + 1])
            for i in range(len(break_list) - 1)
        ]
        dtm_ns['bin_chr'] = pd.cut(
            dtm_ns['value'], break_list, right=False, labels=labels).astype(str)
      else:
        dtm_ns = pd.merge(dtm_ns, break_df, how='left', on='value')

      dtm_ns = dtm_ns[['idx', 'bin_chr', 'value', 'y']]

    new_dtm = pd.concat([dtm_sv, dtm_ns], ignore_index=True)
    dtm = pd.merge(dtm, new_dtm[['idx', 'bin_chr']], on='idx', how='left')
    bin_res = bin_res.copy()
    bin_res['index'] = bin_res.index
    bin_res['bin_chr'] = bin_res['bin']
    dtm = pd.merge(dtm, bin_res[['bin_chr', value]], on='bin_chr', how='left')
    dtm.set_index(dtm['idx'], drop=True, inplace=True)
    variable = dtm['variable'].iloc[0]
    feature_name = '_'.join([variable, value])
    dtm.rename(columns={value: feature_name}, inplace=True)
    return dtm[feature_name]

  def binning_breaks(self, dtm, breaks):
    """按照给定的breaks进行分箱"""
    break_df = self.split_vec_to_df(breaks)

    # binning
    if is_numeric_dtype(dtm['value']):
      break_list = ['-inf'] + list(
          set(break_df.value.tolist()).difference(
              {np.nan, '-inf', 'inf', 'Inf', '-Inf'})) + ['inf']
      break_list = sorted(list(map(float, break_list)))
      labels = [
          '[{},{})'.format(break_list[i], break_list[i + 1])
          for i in range(len(break_list) - 1)
      ]
      bin_chr = pd.cut(dtm['value'], break_list, right=False, labels=labels)

      binning = self.binning(dtm, bin_chr)

      # 不确定经过正常流程是否还会产生空值，故注释掉下面代码，后续测试中如果发现确实
      # 存在空值再行处理。不建议在此处强行合并到missing中。
      # # sort bin
      # binning = pd.merge(
      #     binning.assign(value=lambda x: [
      #         float(re.search(r"^\[(.*),(.*)\)", i).group(2))
      #         if i != 'nan' else np.nan for i in binning['bin']
      #     ]),
      #     break_df.assign(value=lambda x: x.value.astype(float)),
      #     how='left',
      #     on='value').sort_values(by="rowid").reset_index(drop=True)
      # # merge binning and bk_df if nan isin value
      # if break_df['value'].isnull().any():
      #     binning = binning.assign(bin=lambda x:
      #     [i if i != 'nan' else 'missing' for i in x['bin']]) \
      #         .fillna('missing') \
      #         .groupby(['variable', 'rowid']) \
      #         .agg({'bin': lambda x: '%,%'.join(x),
      #               'good': sum,
      #               'bad': sum}) \
      #         .reset_index()
    else:
      dtm = pd.merge(dtm, break_df, how='left', on='value')
      binning = self.binning(dtm, dtm['bin_chr'])
      # 保持分箱顺序与传入参数一致
      binning['bin_chr'] = binning['bin_chr'].astype(
          'category').cat.set_categories(
              breaks, ordered=True)
      binning = binning.sort_values(by='bin_chr').reset_index(drop=True)

    return binning

  def binning_format(self, binning):
    """"""

    def sub0(x):
      """substitute 0"""
      return np.where(x == 0, self.epsilon, x)

    _pattern = re.compile(r"^\[(.*), *(.*)\)((%,%missing)*)")

    def _extract_breaks(x):
      gp23 = _pattern.match(x)
      breaks_string = x if gp23 is None else gp23.group(2)
      return breaks_string

    # yapf: disable
    binning = binning.assign(
      count=lambda x: x['good'] + x['bad'],
      bad_dist=lambda x: sub0(x['bad']) / sub0(x['bad']).sum(),
      good_dist=lambda x: sub0(x['good']) / sub0(x['good']).sum()
    ).assign(
      count_distr=lambda x: x['count'] / x['count'].sum(),
      badprob=lambda x: x['bad'] / x['count'],
      woe=lambda x: np.log(x['good_dist'] / x['bad_dist'])
    ).assign(
      bin_iv=lambda x: (x['good_dist'] - x['bad_dist']) * x['woe']
    ).assign(total_iv=lambda x: x['bin_iv'].sum())
    # yapf: enable

    binning['breaks'] = binning['bin_chr'].apply(_extract_breaks)
    binning['is_special_values'] = binning['is_sv']
    binning['bin'] = binning['bin_chr'].astype('str')

    return binning[[
        'variable', 'bin', 'count', 'count_distr', 'good', 'bad', 'badprob',
        'woe', 'bin_iv', 'total_iv', 'breaks', 'is_special_values'
    ]]


class ComposedWOEBin(WOEBin):
  """
    组合型分箱方法，将多个分箱方法按顺序进行调用。详细说明请见WOEBin。

    Args
        bin_object: 待组合WOEBin子类实例列表
    """

  def __init__(self, bin_objects, **kwargs):
    super().__init__(**kwargs)
    self.__bin_objects = bin_objects

  def woebin(self, dtm, breaks=None):
    for bin_obj in self.__bin_objects:
      breaks = bin_obj.woebin(dtm, breaks)

    return breaks


class InitBin(WOEBin):

  @staticmethod
  def check_empty_bins(dtm, breaks):
    """针对数值型变量检查空分箱"""
    bins = pd.cut(dtm['value'], breaks, right=False)
    bin_sample_count = bins.value_counts()
    if np.any(bin_sample_count == 0):
      bin_sample_count = bin_sample_count[bin_sample_count != 0]
      bin_right = set([
          re.match(r'\[(.+),(.+)\)', i).group(1)
          for i in bin_sample_count.index.astype('str')
      ]).difference({'-inf', 'inf'})
      breaks = sorted(list(map(float, ['-inf'] + list(bin_right) + ['inf'])))
    return breaks


@WOEBinFactory.register('quantile')
class QuantileInitBin(InitBin):
  """
    细分箱之等频分箱。对数值型变量，该分箱方法通过分位数寻找切分点。对类别型变量，直接返回所有
    类别值。

    Args:
        initial_bins: 等频分箱的箱数，默认20
        sig_figs: 切分点有效数字位数，默认4
    """

  def __init__(self, initial_bins=20, sig_figs=4, **kwargs):
    super().__init__(**kwargs)
    self.n_bins = initial_bins
    self.sig_figs = sig_figs

  def woebin(self, dtm, breaks=None):
    if is_numeric_dtype(dtm['value']):  # numeric variable
      xvalue = dtm['value'].astype(float)
      breaks = np.quantile(xvalue, np.linspace(0, 1, self.n_bins + 1))
      breaks = round_(np.unique(breaks), self.sig_figs)
      breaks[0] = -np.inf
      breaks[-1] = np.inf
      breaks = np.unique(breaks)
      breaks = self.check_empty_bins(dtm, breaks)
    else:
      breaks = np.unique(dtm['value'])
    return breaks


@WOEBinFactory.register('hist')
class HistogramInitBin(InitBin):
  """
    细分箱之等宽分箱。
    对类别型变量，直接返回所有类别值。
    对数值型变量，首先排除outlier样本，对剩余样本的range等分成`n_bins`等分。

    Args:
        n_bins: 等宽分箱的箱数。
    """

  @staticmethod
  def _pretty(low, high, n):

    def nice_number(x):
      exp = np.floor(np.log10(abs(x)))
      f = abs(x) / 10**exp
      if f < 1.5:
        nf = 1.
      elif f < 3.:
        nf = 2.
      elif f < 7.:
        nf = 5.
      else:
        nf = 10.
      return np.sign(x) * nf * 10.**exp

    d = abs(nice_number((high - low) / (n - 1)))
    min_x = np.floor(low / d) * d
    max_x = np.ceil(high / d) * d
    return np.arange(min_x, max_x + 0.5 * d, d)

  # @staticmethod
  # def _check_empty_bins(dtm, breaks):
  #     """针对数值型变量检查空分箱"""
  #     bins = pd.cut(dtm['value'], breaks, right=False)
  #     bin_sample_count = bins.value_counts()
  #     if np.any(bin_sample_count == 0):
  #         bin_sample_count = bin_sample_count[bin_sample_count != 0]
  #         bin_right = set([
  #             re.match(r'\[(.+),(.+)\)', i).group(1)
  #             for i in bin_sample_count.index.astype('str')
  #         ]).difference({'-inf', 'inf'})
  #         breaks = sorted(
  #             list(map(float, ['-inf'] + list(bin_right) + ['inf'])))
  #     return breaks

  def woebin(self, dtm, breaks=None):
    if is_numeric_dtype(dtm['value']):  # numeric variable
      xvalue = dtm['value'].astype(float)

      # outlier处理
      iq = xvalue.quantile([0.01, 0.25, 0.75, 0.99])
      iqr = iq[0.75] - iq[0.25]
      if iqr == 0:
        prob_down = 0.01
        prob_up = 0.99
      else:
        prob_down = 0.25
        prob_up = 0.75
      xvalue_rm_outlier = xvalue[(xvalue >= iq[prob_down] - 3 * iqr)
                                 & (xvalue <= iq[prob_up] + 3 * iqr)]

      n_bins = self.n_bins
      len_uniq_x = len(np.unique(xvalue_rm_outlier))
      if len_uniq_x < n_bins:
        n_bins = len_uniq_x
      # initial breaks
      if len_uniq_x == n_bins:
        breaks = np.unique(xvalue_rm_outlier)
      else:
        breaks = self._pretty(
            low=min(xvalue_rm_outlier),
            high=max(xvalue_rm_outlier),
            n=self.n_bins)

      breaks = list(
          filter(lambda x: np.nanmin(xvalue) < x <= np.nanmax(xvalue), breaks))
      breaks = [float('-inf')] + sorted(breaks) + [float('inf')]
      breaks = self.check_empty_bins(dtm, breaks)
    else:
      breaks = np.unique(dtm['value'])
    return breaks

  def __init__(self, initial_bins=20, **kwargs):
    super().__init__(**kwargs)
    self.n_bins = initial_bins


class OptimBinMixin:
  """最优分箱Mixin，提供initial_binning方法，根据细分箱切分点生成分箱统计表"""

  def initial_binning(self, dtm, breaks):
    binning = self.binning_breaks(dtm, breaks)
    binning['count'] = binning['good'] + binning['bad']
    binning['count_distr'] = binning['count'] / binning['count'].sum()

    if not is_numeric_dtype(dtm['value']):
      binning['badprob'] = binning['bad'] / binning['count']
      binning = binning.sort_values(
          by='badprob', ascending=False).reset_index(drop=True)

    return binning


@WOEBinFactory.register(['chi2', 'chimerge'])
class ChiMergeOptimBin(WOEBin, OptimBinMixin):
  """
    ChiMerge最优分箱方法，对相邻分箱进行chi2列联表独立性检验，基于检验的统计量进行分箱合并。

    TODO: 增加分箱单调性约束

    Args
        - bin_num_limit: 分箱数上限，默认5
        - p: 独立性检验显著性，模型0.05
        - count_distr_limit: 最小分箱样本占比，默认0.02
        - ensure_monotonic: 是否要求单调，默认False（暂不支持该功能）
    """

  def __init__(self,
               bin_num_limit=5,
               p=0.05,
               count_distr_limit=0.02,
               ensure_monotonic=False,
               **kwargs):
    super().__init__(**kwargs)
    self.bin_num_limit = bin_num_limit
    self.p = p
    self.count_distr_limit = count_distr_limit
    self.ensure_monotonic = ensure_monotonic
    self.chi2_limit = chi2.isf(p, df=1)

  @staticmethod
  def chi2_stat(binning):
    """计算两分箱之间的Chi2统计量，这里直接使用`scipy.stats.chi2_contingency`函数，
        并且使用 Yate's 连续性修正"""
    binning['good_lag'] = binning['good'].shift(1)
    binning['bad_lag'] = binning['bad'].shift(1)

    def chi2_cont_tbl(arr):
      if np.any(np.isnan(arr)):
        return np.nan
      elif np.any(np.sum(arr, axis=1) == 0) or np.any(np.sum(arr, axis=0) == 0):
        return 0.0
      else:
        return chi2_contingency(arr, correction=True)[0]

    binning['chi2'] = binning.apply(
        lambda x: chi2_cont_tbl([[x['good'], x['bad']],
                                 [x['good_lag'], x['bad_lag']]]),
        axis=1)
    del binning['good_lag']
    del binning['bad_lag']

    return binning

  def woebin(self, dtm, breaks=None):
    assert breaks is not None, f"使用{self.__class__.__name__}类进行分箱，" f"需要传入初始分箱（细分箱）结果"
    binning = self.initial_binning(dtm, breaks)
    binning_chi2 = self.chi2_stat(binning)
    binning_chi2['bin_chr'] = binning_chi2['bin_chr'].astype('str')

    # Start merge loop
    while True:
      min_chi2 = binning_chi2['chi2'].min()
      min_count_distr = binning_chi2['count_distr'].min()
      n_bins = len(binning_chi2)

      if min_chi2 < self.chi2_limit:
        # 分箱坏占比差异不显著
        idx = binning_chi2[binning_chi2['chi2'] == min_chi2].index[0]
      elif min_count_distr < self.count_distr_limit:
        # 分箱占比过少
        idx = binning_chi2[binning_chi2['count_distr'] ==
                           min_count_distr].index[0]
        if idx == 0 or (idx < len(binning_chi2) - 1 and
                        (binning_chi2['chi2'][idx]
                         > binning_chi2['chi2'][idx + 1])):
          idx = idx + 1
      elif n_bins > self.bin_num_limit:
        # 分箱数太多
        idx = binning_chi2[binning_chi2['chi2'] == min_chi2].index[0]
      else:
        # 结束合并操作
        break

      # yapf: disable
      binning_chi2.loc[idx - 1, 'bin_chr'] = '%,%'.join(
        [binning_chi2.loc[idx - 1, 'bin_chr'],
         binning_chi2.loc[idx, 'bin_chr']])
      binning_chi2.loc[idx - 1, 'count'] = (
          binning_chi2.loc[idx - 1, 'count'] +
          binning_chi2.loc[idx, 'count'])
      binning_chi2.loc[idx - 1, 'count_distr'] = (
          binning_chi2.loc[idx - 1, 'count_distr'] +
          binning_chi2.loc[idx, 'count_distr'])
      binning_chi2.loc[idx - 1, 'good'] = (
          binning_chi2.loc[idx - 1, 'good'] +
          binning_chi2.loc[idx, 'good'])
      binning_chi2.loc[idx - 1, 'bad'] = (
          binning_chi2.loc[idx - 1, 'bad'] +
          binning_chi2.loc[idx, 'bad'])
      # yapf: enable

      if is_numeric_dtype(dtm['value']):
        # 数值类型分箱合并
        # [a,b)%,%[b,c) → [a,c)
        binning_chi2['bin_chr'] = binning_chi2['bin_chr'].apply(
            lambda x: re.sub(r',[.\d]+\)%,%\[[.\d]+,', ',', x))

      index = binning_chi2.index.tolist()
      index.remove(idx)
      binning_chi2 = binning_chi2.iloc[
          index,
      ].reset_index(drop=True)
      binning_chi2 = self.chi2_stat(binning_chi2)
    # End of loop

    # 切分点提取
    if is_numeric_dtype(dtm['value']):
      _pattern = re.compile(r"^\[(.*), *(.*)\)")
      breaks = binning_chi2['bin_chr'].apply(lambda x: _pattern.match(x)[2])
      breaks = pd.to_numeric(breaks)
    else:
      breaks = binning_chi2['bin_chr']

    return breaks


@WOEBinFactory.register('tree')
class TreeOptimBin(WOEBin, OptimBinMixin):
  """
    树分箱方法，从细分箱生成的切分点中挑选最优切分点，自顶向下逐步生成分箱树，完成分箱。

    Args
        bin_num_limit: 分箱数上限，默认5
        min_iv_inc: 增加切分点后IV相对增幅最小值，模型0.05
        count_distr_limit: 最小分箱样本占比，默认0.02
        ensure_monotonic: 是否要求严格单调，默认False
    """

  def __init__(self,
               bin_num_limit=5,
               min_iv_inc=0.05,
               count_distr_limit=0.02,
               ensure_monotonic=False,
               **kwargs):
    super().__init__(**kwargs)
    self.bin_num_limit = bin_num_limit
    self.min_iv_inc = min_iv_inc
    self.count_distr_limit = count_distr_limit
    self.ensure_monotonic = ensure_monotonic

  def woebin(self, dtm, breaks=None):
    assert breaks is not None, f"使用{self.__class__.__name__}类进行分箱，" f"需要传入初始分箱（细分箱）结果"
    binning_tree = self.initial_binning(dtm, breaks)
    binning_tree['node_id'] = 0
    binning_tree['cp'] = False  # cut point flag
    binning_tree.loc[len(binning_tree) - 1, 'cp'] = True

    last_iv = 0

    while len(binning_tree['node_id'].unique()) <= self.bin_num_limit:
      cut_idx_iv = {}
      for idx in binning_tree.index[~binning_tree['cp']]:
        new_node_ids = self.node_split(binning_tree['node_id'], idx)
        new_binning = self.merge_binning(binning_tree, new_node_ids)
        if self.ensure_monotonic:
          monotonic_type = monotonic(new_binning['bad_prob'])
          if monotonic_type in ('increasing', 'decreasing'):
            monotonic_constrain = True
          else:
            monotonic_constrain = False
        else:
          monotonic_constrain = True

        if (np.all(new_binning['count_distr'] > self.count_distr_limit) and
            monotonic_constrain):
          # (pandas==2.1.1)
          # FutureWarning: Series.__getitem__ treating keys as positions is deprecated. In a future version,
          # integer keys will always be treated as labels (consistent with DataFrame behavior). To access a
          # value by position, use `ser.iloc[pos]`
          # curr_iv = new_binning['total_iv'][0]
          curr_iv = new_binning['total_iv'].iloc[0]
          if ((curr_iv - last_iv + 1e-8) / (last_iv + 1e-8)) > self.min_iv_inc:
            cut_idx_iv[idx] = curr_iv

      if len(cut_idx_iv) > 0:
        sorted_cut_idx_iv = sorted(cut_idx_iv.items(), key=lambda x: -x[1])
        best_cut_idx = sorted_cut_idx_iv[0][0]
        last_iv = sorted_cut_idx_iv[0][1]
        binning_tree['node_id'] = self.node_split(binning_tree['node_id'],
                                                  best_cut_idx)
        binning_tree.loc[best_cut_idx, 'cp'] = True
      else:
        break

    best_binning = self.merge_binning(binning_tree, binning_tree['node_id'])

    if is_numeric_dtype(dtm['value']):
      best_binning['bin_chr'] = best_binning['bin_chr'].apply(
          lambda x: re.sub(r',[.\d]+\)%,%\[[.\d]+,', ',', x))
      _pattern = re.compile(r"^\[(.*), *(.*)\)")
      breaks = best_binning['bin_chr'].apply(lambda x: _pattern.match(x)[2])
      breaks = pd.to_numeric(breaks)
    else:
      breaks = best_binning['bin_chr']

    return breaks

  def merge_binning(self, binning, node_ids):
    # yapf: disable
    new_binning = binning.groupby([
      'variable',
      node_ids,
    ]).agg(bin_chr=('bin_chr', lambda x: '%,%'.join(x.tolist())),
           count=('count', 'sum'),
           count_distr=('count_distr', 'sum'),
           good=('good', 'sum'),
           bad=('bad', 'sum')
           ).assign(
      bad_prob=lambda x: x['bad'] / x['count'],
      total_iv=lambda x: self.iv(x['good'], x['bad']))
    # yapf: enable

    return new_binning

  @staticmethod
  def node_split(node_ids, idx):
    new_node_ids = np.where(node_ids.index <= idx, node_ids, node_ids + 1)

    return new_node_ids

  def iv(self, good, bad):
    good = np.asarray(good)
    bad = np.asarray(bad)
    # substitute 0 by self.epsilon
    good = np.where(good == 0, self.epsilon, good)
    bad = np.where(bad == 0, self.epsilon, bad)
    good_distr = good / good.sum()
    bad_distr = bad / bad.sum()
    iv = (good_distr - bad_distr) * np.log(good_distr / bad_distr)
    return iv.sum()


@WOEBinFactory.register('rule')
class RuleOptimBin(WOEBin, OptimBinMixin):
  """规则优化分箱算法，用于生成单变量规则。该分箱方式会生成三个分箱（不包含特殊值分箱），
  分别为拒绝分箱、监控分箱、通过分箱。其中拒绝分箱的坏率提升度需大于 `min_lift`，通过
  分箱为紧邻拒绝分箱、占比>5%的样本，其余为通过分箱。建议上游细分箱方法采用 
  `QuantileInitBin`，且`initial_bins > 20`。

  * 假设检验：拒绝分箱坏率显著高于整体坏率（alpha=0.05），不满足时无法分箱
  * 最小命中样本数：拒绝分箱最少样本数，默认不限制，建议设置为50以上，不满足时无法分箱

  Args:
    min_lift: 要求最小风险提升度，默认为 3
    min_hit_samples: 最小命中样本数，默认为 None 代表不限制命中样本数
  """

  def __init__(self,
               min_lift: float = 3,
               min_hit_samples: Optional[int] = None,
               pvalue: float = 0.05,
               eps: float = 1e-8):
    super().__init__()
    self._min_lift = min_lift
    self._min_hit_samples = min_hit_samples or 0
    self._p = pvalue
    self._eps = eps

  def cut_binning(self, binning, idx):
    flag = np.where(binning.index <= idx, 'left', 'right')
    new_binning = binning.groupby([
        'variable',
        flag,
    ]).agg(
        bin_chr=('bin_chr', lambda x: '%,%'.join(x.tolist())),
        count=('count', 'sum'),
        count_distr=('count_distr', 'sum'),
        good=('good', 'sum'),
        bad=('bad', 'sum')).assign(
            bad_prob=lambda x: x['bad'] / x['count'],
            bad_prob_all=lambda x: x['bad'].sum() / x['count'].sum()).assign(
                lift=lambda x: (x['bad_prob'] + self._eps) / x['bad_prob_all'])
    new_binning['foil'] = new_binning['bad'] * np.log2(new_binning['lift'])

    # yapf: disable
    if not (np.any(new_binning['lift'] > self._min_lift)
      and new_binning['count'].min() > self._min_hit_samples
      and fisher_exact(new_binning[['good', 'bad']]).pvalue < self._p):
      new_binning['foil'] = 0
    # yapf: enable

    return new_binning.reset_index(drop=True)

  def woebin(self, dtm, breaks=None):
    assert breaks is not None, f"使用{self.__class__.__name__}类进行分箱，" f"需要传入初始分箱（细分箱）结果"
    binning = self.initial_binning(dtm, breaks)
    if binning.shape[0] < 2:
      return [-np.inf, np.inf]

    # 步骤1：寻找最优切点
    cut_idx_metric = {}
    for idx in range(binning.shape[0] - 2):
      cut_idx_metric[idx] = self.cut_binning(binning, idx)['foil'].max()
    sorted_cut_idx_metric = sorted(cut_idx_metric.items(), key=lambda x: -x[1])
    best_cut_idx = sorted_cut_idx_metric[0][0]
    best_cut_metric = sorted_cut_idx_metric[0][1]

    # 步骤2：设置监控分箱
    if best_cut_metric == 0:
      # 无法找到最优切点
      return [-np.inf, np.inf]
    else:
      new_binning = self.cut_binning(binning, best_cut_idx)
      binning['cum_count_distr'] = binning['count_distr'].cumsum()
      # yapf: disable
      if new_binning['bad_prob'].is_monotonic_decreasing:
        # 坏率下降，拒绝极小值
        reject_ratio = binning['count_distr'].iloc[best_cut_idx]
        monitor_cut_idx = binning.index[
          binning['cum_count_distr'] >= min(reject_ratio + 0.05, 1)].min()
        if np.isnan(monitor_cut_idx):
          monitor_cut_idx = np.inf
      else:
        # 坏率提升，拒绝极大值
        reject_ratio = 1 - binning['cum_count_distr'].iloc[best_cut_idx]
        monitor_cut_idx = binning.index[
          binning['cum_count_distr'] <= max(1 - reject_ratio - 0.05, 0)].max()
        if np.isnan(monitor_cut_idx):
          monitor_cut_idx = -np.inf
      # yapf: enable

    cut_idx = np.sort(
        np.unique([-np.inf, best_cut_idx, monitor_cut_idx, np.inf]))
    binning['grp'] = pd.cut(binning.index, cut_idx)
    best_binning = binning.groupby(['variable', 'grp'], observed=False).agg(
        bin_chr=('bin_chr', lambda x: '%,%'.join(x.tolist())))

    if is_numeric_dtype(dtm['value']):
      best_binning['bin_chr'] = best_binning['bin_chr'].apply(
          lambda x: re.sub(r',[.\d]+\)%,%\[[.\d]+,', ',', x))
      _pattern = re.compile(r"^\[(.*), *(.*)\)")
      breaks = best_binning['bin_chr'].apply(lambda x: _pattern.match(x)[2])
      breaks = pd.to_numeric(breaks)
    else:
      breaks = best_binning['bin_chr']

    return breaks


def woebin_ply(dt, bins, no_cores=None, replace_blank=False, value='woe'):
  """
    将woebin函数返回分箱结果进行应用，与`scorecardpy.woebin_ply`相比，增加参数`value`。
    该参数可选值为['woe', 'index', 'bin']，当 value='woe' 时，将原始值替换为woe值；
    当 value='index' 时，将原始值替换为变量分箱结果数据框中的 index ，即0, 1, 2,...，可用于
    one-hot编码等进一步处理；当 value='bin' 时，返回结果为分箱区间 [a,b) 【数值型变量】或
    a%,%b 【类别型变量】。

    注意，对应不同 value 参数值，返回数据框中变量名后缀会相应改变。假设原始变量名为 'var_123'，
    返回数据框中列名对应为 'var_123_{value}'。

    Args:
        dt: 包含变量原始值的数据框
        bins: woebin分箱结果
        no_cores: 多进程数量
        replace_blank: 是否将空字符串 '' 替换为 np.nan。注意若为 True 本函数非常耗时，
            建议在预处理阶段就完成该操作。保留该参数是为了保持与 scorecardpy 的兼容性。
        value: 返回值类别，可选['woe', 'index', 'bin']。

    Returns:
        pd.DataFrame，包含入参数据框中未替换的所有列，和替换后的变量列。

    """
  # start time
  start_time = time.time()

  # x variables
  x_vars_bin = bins.keys()
  x_vars_dt = dt.columns.tolist()
  x_vars = list(set(x_vars_bin).intersection(x_vars_dt))
  n_x = len(x_vars)
  # initial data set
  dat = dt.loc[:, list(set(x_vars_dt) - set(x_vars))].copy()

  if no_cores is None or no_cores < 1:
    all_cores = mp.cpu_count() - 1
    no_cores = int(np.ceil(n_x / 5 if n_x / 5 < all_cores else all_cores * 0.9))
  no_cores = max(no_cores, 1)

  tasks = [
      (
          pd.DataFrame({
              'y': 0,  # 不重要
              'variable': var,
              'value': dt[var]
          }),
          bins[var],
          value) for var in x_vars
  ]

  if no_cores == 1:
    dat_suffix = list(itertools.starmap(WOEBin.apply, tasks))
  else:
    pool = mp.Pool(processes=no_cores)
    dat_suffix = pool.starmap(WOEBin.apply, tasks)
    pool.close()

  dat = pd.concat([dat] + dat_suffix, axis=1)

  # running time
  running_time = time.time() - start_time
  logging.info('Woe transformation on {} rows and {} columns in {}'.format(
      dt.shape[0], n_x, time.strftime("%H:%M:%S", time.gmtime(running_time))))
  return dat


def plot_bin(binx, title, show_iv):
  y_right_max = np.ceil(binx['badprob'].max() * 10)
  if y_right_max % 2 == 1:
    y_right_max = y_right_max + 1
  if y_right_max - binx['badprob'].max() * 10 <= 0.3:
    y_right_max = y_right_max + 2
  y_right_max = y_right_max / 10
  if (y_right_max > 1 or y_right_max <= 0 or y_right_max is np.nan or
      y_right_max is None):
    y_right_max = 1
  # y_left_max
  y_left_max = np.ceil(binx['count_distr'].max() * 10) / 10
  if (y_left_max > 1 or y_left_max <= 0 or y_left_max is np.nan or
      y_left_max is None):
    y_left_max = 1
  # title
  title_string = binx.loc[0, 'variable'] + "  (iv:" + str(
      round(binx.loc[0,
                     'total_iv'], 4)) + ")" if show_iv else binx.loc[0,
                                                                     'variable']
  title_string = (
      title + '-' + title_string if title is not None else title_string)
  # param
  ind = np.arange(len(binx.index))  # the x locations for the groups
  width = 0.35  # the width of the bars: can also be len(x) sequence
  # plot
  fig, ax1 = plt.subplots()
  ax2 = ax1.twinx()
  # ax1
  p1 = ax1.bar(
      ind, binx['good_distr'], width, color=(24 / 254, 192 / 254, 196 / 254))
  p2 = ax1.bar(
      ind,
      binx['bad_distr'],
      width,
      bottom=binx['good_distr'],
      color=(246 / 254, 115 / 254, 109 / 254))
  for i in ind:
    ax1.text(
        i,
        binx.loc[i, 'count_distr'] * 1.02,
        str(round(binx.loc[i, 'count_distr'] * 100, 1)) + '%, ' +
        str(binx.loc[i, 'count']),
        ha='center')
  # ax2
  ax2.plot(ind, binx['badprob'], marker='o', color='blue')
  for i in ind:
    ax2.text(
        i,
        binx.loc[i, 'badprob'] * 1.02,
        str(round(binx.loc[i, 'badprob'] * 100, 1)) + '%',
        color='blue',
        ha='center')
  # settings
  ax1.set_ylabel('Bin count distribution')
  ax2.set_ylabel('Bad probability', color='blue')
  ax1.set_yticks(np.arange(0, y_left_max + 0.2, 0.2))
  ax2.set_yticks(np.arange(0, y_right_max + 0.2, 0.2))
  ax2.tick_params(axis='y', colors='blue')
  plt.xticks(ind, binx['bin'])
  plt.title(title_string, loc='left')
  plt.legend((p2[0], p1[0]), ('bad', 'good'), loc='upper right')

  return fig


def woebin_plot(bins, x=None, title=None, show_iv=True):
  xs = x
  # bins concat
  if isinstance(bins, dict):
    bins = pd.concat(bins, ignore_index=True)

  # good bad distr
  def gb_distr(bin_x):
    bin_x['good_distr'] = bin_x['good'] / sum(bin_x['count'])
    bin_x['bad_distr'] = bin_x['bad'] / sum(bin_x['count'])
    return bin_x

  bins = bins.groupby('variable').apply(gb_distr)
  # x variable names
  if xs is None:
    xs = bins['variable'].unique()
  # plot export
  plot_list = {}
  for i in xs:
    binx = bins[bins['variable'] == i].reset_index()
    plot_list[i] = plot_bin(binx, title, show_iv)
  return plot_list


def sc_bins_to_df(sc_bins):
  """
    将 woebin 返回的结果转换为 woe 数据框、iv 数据框
    Args:
        sc_bins: dict, 由 woebin 返回，结构为 {'VAR_NAME': 'BIN_STATS'}

    Returns:
        (woe_df, iv_df)

    """
  woe_df = None
  for key, value in sc_bins.items():
    if isinstance(value, pd.DataFrame):
      if woe_df is None:
        woe_df = value
      else:
        woe_df = pd.concat([woe_df, value], axis=0, ignore_index=True)

  def iv_stats(x):
    iv = x.total_iv.max()
    badrate = x['bad'].sum() / x['count'].sum()
    lift = x.badprob / badrate
    iv_interval = None
    if iv < 0.02:
      iv_interval = '(0, 0.02)'
    elif iv < 0.05:
      iv_interval = '[0.02, 0.05)'
    elif iv < 0.08:
      iv_interval = '[0.05, 0.08)'
    elif iv < 0.1:
      iv_interval = '[0.08, 0.1)'
    elif iv < 0.2:
      iv_interval = '[0.1, 0.2)'
    else:
      iv_interval = '[0.2, +)'

    badrate = x[~x.is_special_values].badprob
    monotonic_type = monotonic(badrate)

    return pd.Series(
        [iv, iv_interval, monotonic_type,
         lift.max(), lift.min()],
        index=['IV', 'IV区间', '单调性', '最大Lift', '最小Lift'],
        dtype='object')

  if woe_df is None:
    return None, None
  else:
    iv_df = woe_df.groupby(by='variable').apply(iv_stats, include_groups=False)
    iv_df.sort_values(by='IV', ascending=False, inplace=True)
    return woe_df, iv_df


def make_scorecard(sc_bins, coef, *, base_points=600, base_odds=50, pdo=20):
  a = pdo / np.log(2)
  b = base_points - a * np.log(base_odds)

  base_score = -a * coef['const'] + b
  score_df = [
      pd.DataFrame({
          'variable': ['base score'],
          'bin': [''],
          'woe': [''],
          'score': [base_score]
      })
  ]

  for var in coef.keys():
    if var != 'const':
      woe_df = sc_bins[var[:-4]][['variable', 'bin', 'woe']].copy()
      woe_df['score'] = -a * coef[var] * woe_df['woe']
      score_df.append(woe_df)

  score_df = pd.concat(score_df, ignore_index=True)
  score_df['score'] = np.round(score_df['score'], 2)
  return score_df


def woebin_breaks(bins):
  """
    从woebin返回结果中提取切分点及特殊值
    Args:
        bins: woebin函数的返回结果

    Returns:
        breaks字典和special_values字典组成的元组

    """

  def get_breaks(binning):
    if np.any(binning['is_special_values']):
      special_values = binning[binning['is_special_values']]['breaks']
      special_values = special_values.tolist()

      # 注意，此处需要将 missing 从 special_values 中排除。否则当 dtm 无空值
      # 且 value 为数值型，而 special_values 中存在 missing 时，
      # WOEBin.split_special_values 中 merge 步会报如下错误：
      # ValueError: You are trying to merge on float64 and object columns.
      # If you wish to proceed you should use pd.concat

      if 'missing' in special_values:
        special_values.remove('missing')
      if len(special_values) == 0:
        special_values = None

    else:
      special_values = None

    breaks = binning[~binning['is_special_values']]['breaks']
    breaks = breaks.tolist()
    return {'breaks': breaks, 'special_values': special_values}

  brk_spcs = {key: get_breaks(value) for key, value in bins.items()}
  breaks = {k: v['breaks'] for k, v in brk_spcs.items()}
  special_values = {
      k: v['special_values'] for k, v in brk_spcs.items() if v['special_values']
  }
  return breaks, special_values


def woebin_psi(df_base, df_cmp, bins):
  """
    计算变量PSI

    Args:
        df_base: 基准数据集，一般为训练集
        df_cmp: 比较数据集，一般为测试集、OOT等等
        bins: 变量分箱结果，由`woebin`返回

    Returns:
        pd.DataFrame，包含variable, bin, base%, cmp%, total_psi五列

    """

  # replace_blank is very expensive, so set the arg to False. But this may
  # cause errors. A good practice is replacing the blank str to np.nan in
  # data preprocessing procedure.
  X_base = woebin_ply(df_base, bins, value='bin', replace_blank=False)
  X_cmp = woebin_ply(df_cmp, bins, value='bin', replace_blank=False)

  vars_base = [v for v in X_base.columns if v.endswith('_bin')]
  vars_cmp = [v for v in X_cmp.columns if v.endswith('_bin')]
  variables = list(set(vars_base).intersection(set(vars_cmp)))

  X_base['set'] = 'base'
  X_cmp['set'] = 'cmp'

  dat = pd.concat([X_base, X_cmp])
  dat['idx'] = dat.index

  psi_dfs = []

  for variable in variables:
    psi_df = pd.pivot_table(
        dat, index=variable, columns=['set'], values=['idx'], aggfunc='count')
    psi_df.columns = ['base', 'cmp']
    psi_df['variable'] = variable[:-4]
    psi_df['bin'] = psi_df.index
    psi_df.reset_index(drop=True, inplace=True)

    psi_df = psi_df.assign(
        base_distr=lambda x: x['base'] / x['base'].sum(),
        cmp_distr=lambda x: x['cmp'] / x['cmp'].sum()).assign(
            psi=lambda x: psi(x['base_distr'], x['cmp_distr']))[[
                'variable', 'bin', 'base_distr', 'cmp_distr', 'psi'
            ]]
    psi_dfs.append(psi_df)

  return pd.concat(psi_dfs, ignore_index=True)



#!/usr/bin/python
"""
数据探索，包括变量类型、空值率、分布等统计类数据
本模块来自 toad (https://github.com/amphibian-dev/toad)
"""
from concurrent.futures import ProcessPoolExecutor, as_completed
try:
  from types import NoneType
except (ModuleNotFoundError, ImportError):
  # Python<=3.8 has not types module
  NoneType = type(None)
from typing import Union

import pandas as pd


def getTopValues(series, top=5, reverse=False):
  """Get top/bottom n values
    Args:
        series (Series): data series
        top (number): number of top/bottom n values
        reverse (bool): it will return bottom n values if True is given
    Returns:
        Series: Series of top/bottom n values and percentage. ['value:percent', None]
    """
  itype = 'top'
  counts = series.value_counts()
  counts = list(zip(counts.index, counts, counts.divide(series.size)))

  if reverse:
    counts.reverse()
    itype = 'bottom'

  template = "{0[0]}:{0[2]:.2%}"
  indexs = [itype + str(i + 1) for i in range(top)]
  values = [
      template.format(counts[i]) if i < len(counts) else None
      for i in range(top)
  ]

  return pd.Series(values, index=indexs)


def getDescribe(series, percentiles=None):
  """Get describe of series
    Args:
        series (Series): data series
        percentiles: the percentiles to include in the output
    Returns:
        Series: the describe of data include mean, std, min, max and percentiles
    """
  if percentiles is None:
    percentiles = [.25, .5, .75]
  d = series.describe(percentiles)
  return d.drop('count')


def countBlank(series, blanks=None):
  """Count number and percentage of blank values in series
    Args:
        series (Series): data series
        blanks (list): list of blank values
    Returns:
        number: number of blanks
        str: the percentage of blank values
    """
  if blanks is None:
    blanks = []

  if len(blanks) > 0:
    isnull = series.replace(blanks, None).isnull()
  else:
    isnull = series.isnull()
  n = isnull.sum()
  ratio = isnull.mean()

  return n, "{0:.2%}".format(ratio)


def isNumeric(series):
  """Check if the series' type is numeric
    Args:
        series (Series): data series
    Returns:
        bool
    """
  return series.dtype.kind in 'ifc'


def _detect(name, series):
  numeric_index = [
      'mean', 'std', 'min', '1%', '10%', '50%', '75%', '90%', '99%', 'max'
  ]
  discrete_index = [
      'top1', 'top2', 'top3', 'top4', 'top5', 'bottom5', 'bottom4', 'bottom3',
      'bottom2', 'bottom1'
  ]

  details_index = [
      numeric_index[i] + '_or_' + discrete_index[i]
      for i in range(len(numeric_index))
  ]
  details = []

  if isNumeric(series):
    desc = getDescribe(series, percentiles=[.01, .1, .5, .75, .9, .99])
    details = desc.tolist()
  else:
    top5 = getTopValues(series)
    bottom5 = getTopValues(series, reverse=True)
    details = top5.tolist() + bottom5[::-1].tolist()

  # print(details_index)
  nblank, pblank = countBlank(series)

  row = pd.Series(
      index=['type', 'size', 'missing', 'unique'] + details_index,
      data=[series.dtype, series.size, pblank,
            series.nunique()] + details)

  row.name = name
  return row


def detect(dataframe, n_cores: Union[NoneType, int] = None):
  """ Detect data
    Args:
        dataframe (DataFrame): data that will be detected
        n_cores (int): parallel computing if n_cores is None or n_cores > 1.
    Returns:
        DataFrame: report of detecting
    """

  rows = []
  # n_cores = int(n_cores) if n_cores >= 1 else None
  n_cores = None if n_cores is None or n_cores < 1 else int(n_cores)
  n_cols = dataframe.shape[1]
  if n_cols < 10 or n_cores == 1:
    for name, series in dataframe.items():
      row = _detect(name, series)
      rows.append(row)
  else:
    with ProcessPoolExecutor(max_workers=n_cores) as executor:
      futures = [
          executor.submit(_detect, name, series)
          for name, series in dataframe.items()
      ]
      for fut in as_completed(futures):
        rows.append(fut.result())

  return pd.DataFrame(rows)



# -*- encoding: utf-8 -*-

import unittest

import numpy as np
import pandas as pd

from addpath import addpath

try:
  addpath()
  from syriskmodels.scorecard import WOEBinFactory
  from syriskmodels.scorecard import WOEBin
  from syriskmodels.scorecard import woebin
  from syriskmodels.contrib.build_scorecard import build_scorecard
  from syriskmodels.scorecard import RuleOptimBin
  from syriskmodels.scorecard import QuantileInitBin
  from syriskmodels.scorecard import sc_bins_to_df
except ModuleNotFoundError as e:
  raise e


class TestRuleOptimBin(unittest.TestCase):

  def setUp(self) -> None:
    self.df: pd.DataFrame = pd.read_hdf('dataset.h5', key='germancredit')
    self.df2: pd.DataFrame = pd.read_hdf('dataset.h5', key='creditcard')
    pd.set_option('display.max_columns', 100)

  def test_rulebin(self):
    dtm = self.df2[['V3', 'Class']].copy()
    dtm.rename(columns={'V3': 'value', 'Class': 'y'}, inplace=True)
    dtm['variable'] = 'V3'
    q_bin = QuantileInitBin(initial_bins=50)
    breaks = q_bin.woebin(dtm)
    r_bin = RuleOptimBin()
    print(r_bin.woebin(dtm, breaks))

    print(
        woebin(
            self.df2,
            x=['V3'],
            y='Class',
            methods=[QuantileInitBin(initial_bins=50),
                     RuleOptimBin()])['V3'])

  def test_rulebin2(self):
    dtm = self.df2[['V4', 'Class']].copy()
    dtm.rename(columns={'V4': 'value', 'Class': 'y'}, inplace=True)
    dtm['variable'] = 'V4'
    q_bin = QuantileInitBin(initial_bins=50)
    breaks = q_bin.woebin(dtm)
    r_bin = RuleOptimBin()
    print(r_bin.woebin(dtm, breaks))

    print(
        woebin(
            self.df2,
            x=['V4'],
            y='Class',
            methods=[QuantileInitBin(initial_bins=50),
                     RuleOptimBin()])['V4'])

  def test_rulebin3(self):
    variables = ['V' + str(i) for i in range(1, 29)]
    bins = woebin(
        self.df2,
        x=variables,
        y='Class',
        methods=[QuantileInitBin(50), RuleOptimBin()])
    woe, _ = sc_bins_to_df(bins)
    woe.to_csv('rule_bin.csv', index=False)


class TestWOEBin(unittest.TestCase):

  def setUp(self) -> None:
    self.df = pd.read_hdf('dataset.h5', key='germancredit')
    self.df2 = pd.read_hdf('dataset.h5', key='creditcard')

  def test_split_vec_to_df(self):
    x = ['b%,%d', 'a', 'c%,%e']
    df = WOEBin.split_vec_to_df(x)
    print(df)

  def test_woebin_cat_vars(self):
    tmp_df = pd.DataFrame({
        'variable': 'property',
        'value': self.df['property'],
        'y': np.where(self.df['creditability'] == 'good', 0, 1)
    })
    woe_bin_method = WOEBinFactory.build(['quantile', 'tree'])
    binning_result = woe_bin_method(tmp_df)
    print(binning_result)

  def test_chi2_woebin(self):
    binner = WOEBinFactory.build(['quantile', 'chi2'])
    tmp_df = pd.DataFrame({
        'variable': 'V3',
        'value': self.df2['V3'],
        'y': self.df2['Class']
    })
    binning_result = binner(tmp_df)
    print(binning_result)

  def test_build_scorecard(self):
    features = self.df2.columns.tolist()[1:-1]
    print(np.quantile(self.df2['Time'], q=0.8))
    build_scorecard(
        self.df2,
        features=features,
        target='Class',
        train_filter=lambda x: x['Time'] <= 140000,
        oot_filter=lambda x: x['Time'] > 140000,
        output_excel_file='test.xlsx')


if __name__ == '__main__':
  unittest.main()
